"""
A bit on semantics:
- blocks are parts of the data (individual files in a file list, parts of a data array)
- individual blocks are assigned specific groups (train/test/val)
"""
from typing import Optional, Dict, List, Any, Tuple, TypeVar
import numpy as np
import pandas as pd

Block = TypeVar("Block")
Group = TypeVar("Group")


def groupstats(stats: List, groups: List[Group]):
    """_summary_

    Args:
        stats (_type_): stats [x] for each group [n, x]
        groups (_type_): ids for each group [n,]

    Returns:
        Dict[group_id, x]: stats average over all members of groups
    """
    unique_groups, group_index = np.unique(groups, return_inverse=True)

    group_stats = {}
    for index, group in enumerate(unique_groups):
        group_stats[group] = np.mean(stats[group_index == index, :], axis=0)
    return group_stats


def group_splits(data: np.ndarray, group_sizes: List[str]) -> List[np.ndarray]:
    group_cdf = np.insert(np.cumsum(group_sizes), 0, 0)[:-1]
    split_points = (group_cdf * data.shape[0]).astype(int)
    blocks = np.array_split(data, split_points[1:])

    split_points = list(split_points)
    split_points.append(data.shape[0])
    split_points = [(start, end) for start, end in zip(split_points[:-1], split_points[1:])]

    return blocks, split_points


def group_blocks(nb_blocks: int, group_sizes: List[str], group_names: List[Group]):
    group_cdf = np.insert(np.cumsum(group_sizes), 0, 0)[:-1]
    group_probs = np.linspace(0, 1, nb_blocks)
    groups = np.zeros((nb_blocks,), dtype=int)
    for cnt, s in enumerate(group_cdf):
        groups[group_probs >= s] = cnt

    groups = [group_names[s] for s in groups]
    return groups


def score_grouping(block_stats, groups: List[Group]):

    # compute global probs
    # we want the individual groups to have stats as close as possible to the global stats
    global_stats = np.mean(block_stats, axis=0)
    # exclude groups with zero occurrence from score
    non_zero_stats = np.where(global_stats > 0)[0]

    # score each group
    group_stats = groupstats(block_stats, groups)
    group_scores = {}
    for group_name, group_stat in group_stats.items():
        tmp = np.log2((group_stat[non_zero_stats] + 0.0000001) / global_stats[non_zero_stats])
        group_scores[group_name] = np.sum(np.abs(tmp))

    # compute total score over all groups
    total_score = np.sum(list(group_scores.values()))
    return total_score


def opt_grouping(
    block_stats: np.ndarray, group_sizes: List[Group], group_names: List[Group], nb_perms: int = 100
) -> List[Group]:

    nb_blocks = len(block_stats)
    grouping = group_blocks(nb_blocks, group_sizes, group_names)
    grouping = np.array(grouping)

    best_perm = []
    best_score = np.inf
    for _ in range(nb_perms):
        perm = np.random.permutation(grouping)
        score = score_grouping(block_stats, grouping)

        if score < best_score:
            best_score = score
            best_perm = perm

    return best_perm


def block_data(data: np.ndarray, block_size: int):
    split_points = np.array(range(0, data.shape[0], block_size))
    split_points[-1] = data.shape[0]
    # Returns a view so free in terms of memory.
    blocks = np.array_split(data, split_points[1:])[:-1]

    # ALTERNATIVE: block_size will be chosen so that all blocks have roughly equal size
    # blocks = np.array_split(y, np.floor(data.shape[0] / block_size))

    # DEBUG CODE
    # print([block.shape for block in blocks])
    # print(len(blocks), blocks[1].shape, blocks[-1].shape, np.sum([len(b) for b in blocks]))
    return blocks, split_points


def blocks_from_split_points(data: np.ndarray, split_points: List[Tuple[int, int]]):
    blocks = []
    for split_point in split_points:
        blocks.append(data[split_point[0] : split_point[1]])
    return blocks


def blockstats_from_data(data: np.ndarray, block_size: int, gap: int = 0) -> Dict[Tuple[int, int], np.ndarray]:
    """_summary_

    Last block will contain overhang so is typically longer

    Args:
        data (np.ndarray): _description_
        block_size (int): _description_
        gap (int, optional)

    Returns:
        Dict[Tuple[int, int], np.array]: _description_
    """
    blocks, split_points = block_data(data, block_size)

    blockstats = {}
    for split_start, split_end, block in zip(split_points[:-1], split_points[1:], blocks):
        blockstats[(split_start, split_end)] = np.mean(block[gap : block.shape[0] - gap], axis=0)
    return blockstats


def blockstats_from_files(file_bases: List[str], class_names: Optional[List] = None) -> Dict[str, np.ndarray]:
    """_summary_

    Args:
        file_bases (List[str]): List of annotation files.
        class_names (Optional[List]): List of class_names. Defaults to None (will infer from all annotations).

    Returns:
        Dict[str, np.ndarray]: Block stats for each file name.
    """
    if class_names is None:
        class_names = []
        for file_base in file_bases:
            df = pd.read_csv(file_base)
            df = df.dropna()
            class_names.extend(list(set(df["name"])))
        class_names = list(set(class_names))

    stratify_by_number = dict()
    for file_base in file_bases:
        df = pd.read_csv(file_base)
        df = df.dropna()
        # stratify_by_number[file_base] = {n: df["name"].tolist().count(n) for n in class_names}
        stratify_by_number[file_base] = [df["name"].tolist().count(n) for n in class_names]

    return stratify_by_number


def format_by_group(groups, block_names, group_names):
    blocks = {group_name: [] for group_name in group_names}
    for block_name, group in zip(block_names, groups):
        blocks[group].append(block_name)
    return blocks


def format_by_block(blocks, block_names):
    block_dict = dict()
    for block_name, block in zip(block_names, blocks):
        block_dict[block_name] = block
    return block_dict


def block(
    block_names: List[np.ndarray],
    group_sizes: List[float],
    group_names: List[Group],
    block_stats: Optional[List[Block]] = None,
    shuffle: bool = True,
    seed: Optional[float] = None,
) -> Dict[Group, List[Block]]:

    if seed is not None:
        np.random.seed(seed)

    if block_stats is None:
        blocks = group_blocks(len(block_names), group_sizes, group_names)
        if shuffle:
            blocks = np.random.permutation(blocks)
    else:
        block_stats = np.array(list(block_stats))
        blocks = opt_grouping(block_stats, group_sizes, group_names)

    blocks = format_by_block(blocks, block_names)

    return blocks



"""

Kapre backend functions
=======================\

|  Some backend functions that mainly use numpy.
|  Functions with Keras' backend is in ``backend_keras.py``.

Notes
-----
    * Don't forget to use ``K.float()``! Otherwise numpy uses float64.
    * Some functions are copied-and-pasted from librosa (to reduce dependency), but
        later I realised it'd be better to just use it.
    * TODO: remove copied code and use librosa.
"""
from tensorflow.keras import backend as K
import numpy as np
import librosa

# Forward compatability to replace xrange
from builtins import range

EPS = 1e-7


def eps():
    return EPS


def mel(sr, n_dft, n_mels=128, fmin=0.0, fmax=None, htk=False, norm=1):
    """[np] create a filterbank matrix to combine stft bins into mel-frequency bins
    use Slaney (said Librosa)

    n_mels: numbre of mel bands
    fmin : lowest frequency [Hz]
    fmax : highest frequency [Hz]
        If `None`, use `sr / 2.0`
    """
    return librosa.filters.mel(sr=sr, n_fft=n_dft, n_mels=n_mels, fmin=fmin, fmax=fmax, htk=htk, norm=norm).astype(K.floatx())


def get_stft_kernels(n_dft):
    """[np] Return dft kernels for real/imagnary parts assuming
        the input . is real.
    An asymmetric hann window is used (scipy.signal.hann).

    Parameters
    ----------
    n_dft : int > 0 and power of 2 [scalar]
        Number of dft components.

    Returns
    -------
        |  dft_real_kernels : np.ndarray [shape=(nb_filter, 1, 1, n_win)]
        |  dft_imag_kernels : np.ndarray [shape=(nb_filter, 1, 1, n_win)]

    * nb_filter = n_dft/2 + 1
    * n_win = n_dft

    """
    assert n_dft > 1 and ((n_dft & (n_dft - 1)) == 0), "n_dft should be > 1 and power of 2, but n_dft == %d" % n_dft

    nb_filter = int(n_dft // 2 + 1)

    # prepare DFT filters
    timesteps = np.array(range(n_dft))
    w_ks = np.arange(nb_filter) * 2 * np.pi / float(n_dft)
    dft_real_kernels = np.cos(w_ks.reshape(-1, 1) * timesteps.reshape(1, -1))
    dft_imag_kernels = -np.sin(w_ks.reshape(-1, 1) * timesteps.reshape(1, -1))

    # windowing DFT filters
    dft_window = librosa.filters.get_window("hann", n_dft, fftbins=True)  # _hann(n_dft, sym=False)
    dft_window = dft_window.astype(K.floatx())
    dft_window = dft_window.reshape((1, -1))
    dft_real_kernels = np.multiply(dft_real_kernels, dft_window)
    dft_imag_kernels = np.multiply(dft_imag_kernels, dft_window)

    dft_real_kernels = dft_real_kernels.transpose()
    dft_imag_kernels = dft_imag_kernels.transpose()
    dft_real_kernels = dft_real_kernels[:, np.newaxis, np.newaxis, :]
    dft_imag_kernels = dft_imag_kernels[:, np.newaxis, np.newaxis, :]

    return dft_real_kernels.astype(K.floatx()), dft_imag_kernels.astype(K.floatx())


def filterbank_mel(sr, n_freq, n_mels=128, fmin=0.0, fmax=None, htk=False, norm=1):
    """[np]"""
    return mel(sr, (n_freq - 1) * 2, n_mels=n_mels, fmin=fmin, fmax=fmax, htk=htk, norm=norm).astype(K.floatx())


def filterbank_log(sr, n_freq, n_bins=84, bins_per_octave=12, fmin=None, spread=0.125):  # pragma: no cover
    """[np] Approximate a constant-Q filter bank for a fixed-window STFT.

    Each filter is a log-normal window centered at the corresponding frequency.

    Note: `logfrequency` in librosa 0.4 (deprecated), so copy-and-pasted,
        `tuning` was removed, `n_freq` instead of `n_fft`.

    Parameters
    ----------
    sr : number > 0 [scalar]
        audio sampling rate

    n_freq : int > 0 [scalar]
        number of frequency bins

    n_bins : int > 0 [scalar]
        Number of bins.  Defaults to 84 (7 octaves).

    bins_per_octave : int > 0 [scalar]
        Number of bins per octave. Defaults to 12 (semitones).

    fmin : float > 0 [scalar]
        Minimum frequency bin. Defaults to `C1 ~= 32.70`

    spread : float > 0 [scalar]
        Spread of each filter, as a fraction of a bin.

    Returns
    -------
    C : np.ndarray [shape=(n_bins, 1 + n_fft/2)]
        log-frequency filter bank.
    """

    if fmin is None:
        fmin = 32.70319566

    # What's the shape parameter for our log-normal filters?
    sigma = float(spread) / bins_per_octave

    # Construct the output matrix
    basis = np.zeros((n_bins, n_freq))

    # Get log frequencies of bins
    log_freqs = np.log2(librosa.fft_frequencies(sr, (n_freq - 1) * 2)[1:])

    for i in range(n_bins):
        # What's the center (median) frequency of this filter?
        c_freq = fmin * (2.0 ** (float(i) / bins_per_octave))

        # Place a log-normal window around c_freq
        basis[i, 1:] = np.exp(-0.5 * ((log_freqs - np.log2(c_freq)) / sigma) ** 2 - np.log2(sigma) - log_freqs)

    # Normalize the filters
    basis = librosa.util.normalize(basis, norm=1, axis=1)

    return basis.astype(K.floatx())



"""Dict of dicts <-> hierarchy of npy files.

Tools for working with dictionaries of dictionaries.
The keys of the first dictionary are mapped to directories.
The values in the nested dictionary are saved as npy files with the key as the name.

As an example, this code:
```python
data = {'first_level': {'song': some_data, 'response': some_other_data}}
data.attrs = {'song_name': 'this is the dream...', 'response_name': 'yeah'}
das.npy.save('song_responses.npy', data)
```
will result in a folder `song_responses.npy` containing
one subfolder `first_level` with two npy files - `song.npy` containing `some_data` and
`responses.npy` containing `some_other_data`. The `attrs` attribute can be used to store metadata and will be stored in `song_responses.npy/attrs.npy`)

`data = das.npy.load('song_responses.npy')` will restore the data with the metadata.
"""
import numpy as np
import os
import os.path
from glob import glob
from typing import Any, List, Dict, Union, Optional


class DictClass(dict):
    """Wrap dict in class so we can attach attrs to it."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attrs: Dict = {}

    def __str__(self):
        out = f"Data:\n"
        for top_key in self.keys():
            out = out + f"   {top_key}:\n"
            for key, val in self[top_key].items():
                out = out + f"      {key}: {val.shape}\n"
        out = out + f"\nAttributes:\n"
        for key, val in self.attrs.items():
            out = out + f"    {key}: {val}\n"
        return out


def load(location: str, memmap_dirs: Optional[Union[List[str], str]] = None) -> DictClass:
    """Load hierarchy of npy files into dict of dicts.

    Args:
        location ([type]): [description]
        memmap_dirs (list, optional): List of dirs to memmap. String 'all' will memmmap all dirs. Defaults to ['train'].

    Returns:
        dict: [description]
    """

    if memmap_dirs is None:
        memmap_dirs = ["train"]

    def path_to_key(path):
        key = os.path.splitext(os.path.basename(path))[0]
        return key

    dir_names = [os.path.join(location, name) for name in os.listdir(location) if os.path.isdir(os.path.join(location, name))]
    data = DictClass()

    attrs_path = os.path.join(location, "attrs.npy")
    if os.path.exists(attrs_path):
        data.attrs = np.load(attrs_path, allow_pickle=True)[()]

    for dir_name in dir_names:
        dir_key = path_to_key(dir_name)
        data[dir_key] = dict()
        npy_files = glob(os.path.join(dir_name, "*.npy"))
        for npy_file in npy_files:
            npy_key = path_to_key(npy_file)
            if memmap_dirs == "all" or dir_key in memmap_dirs:
                data[dir_key][npy_key] = np.lib.format.open_memmap(npy_file, "r")
            else:
                data[dir_key][npy_key] = np.load(npy_file)
    return data


def save(location: str, data: DictClass) -> None:
    """Save nested dict in data to location as a directory with npy files dir.

    Args:
        location ([type]): [description]
        data ([type]): [description]
    """
    os.makedirs(location, exist_ok=True)
    if hasattr(data, "attrs"):
        np.save(os.path.join(location, "attrs"), dict(data.attrs), allow_pickle=True)

    for key_top in data.keys():
        os.makedirs(os.path.join(location, key_top), exist_ok=True)
        for key, val in data[key_top].items():
            np.save(os.path.join(location, key_top, key + ".npy"), val)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import numpy as np
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer
from . import backend, backend_keras
from typing import Optional


class Spectrogram(Layer):
    """Spectrogram layer returns spectrogram(s).

    Examples:

        >>> kapre.time_frequency.Spectrogram(
                    n_dft=512, n_hop=None, padding='same',
                    power_spectrogram=2.0, return_decibel_spectrogram=False,
                    trainable_kernel=False, image_data_format='default'
                    )
    """

    def __init__(
        self,
        n_dft: int = 512,
        n_hop: Optional[int] = None,
        padding: str = "same",
        power_spectrogram: float = 2.0,
        return_decibel_spectrogram: bool = False,
        trainable_kernel: bool = False,
        image_data_format: str = "default",
        **kwargs,
    ) -> Layer:
        """[summary]

        Args:
            n_dft (int, optional): The number of DFT points. Best if power of 2. Defaults to 512.
            n_hop (Optional[int], optional): Hop length between frames in sample. Best if <= `n_dft`. Defaults to None.
            padding (str, optional): Pads signal boundaries (`same` or `valid`). Defaults to 'same'.
            power_spectrogram (float, optional): `2.0` for power, `1.0` for amplitude spectrogram. Defaults to 2.0 (power).
            return_decibel_spectrogram (bool, optional): Convert spectrogram values to dB. Recommended. Defaults to False.
            trainable_kernel (bool, optional): If True, kernels will be optimized during training. Defaults to False.
            image_data_format (str, optional): `channels_first` or `channels_last` or keras' `default`. Defaults to 'default'.

        Notes:
            - The input should be a 2D array, `(audio_channel, audio_length)`. E.g., `(1, 44100)` for mono signals, `(2, 44100)` for stereo signals.
            - Supports multichannel inputs, so `audio_channel` can be any positive integer.
            - The input shape is not related to keras `image_data_format()` config.

        Returns:
            Layer: Keras layer computing the spectrogram
                    - if `channels_first`: `(None, n_channel, n_time, n_freq, )`
                    - if `channels_last`: `(None, n_time, n_freq, n_channel)`
        """
        assert n_dft > 1 and ((n_dft & (n_dft - 1)) == 0), "n_dft should be > 1 and power of 2, but n_dft == %d" % n_dft
        assert isinstance(trainable_kernel, bool)
        assert isinstance(return_decibel_spectrogram, bool)
        assert padding in ("same", "valid")
        if n_hop is None:
            n_hop = n_dft // 2

        assert image_data_format in ("default", "channels_first", "channels_last")
        if image_data_format == "default":
            self.image_data_format = K.image_data_format()
        else:
            self.image_data_format = image_data_format

        self.n_dft = n_dft
        assert n_dft % 2 == 0
        self.n_filter = n_dft // 2 + 1
        self.trainable_kernel = trainable_kernel
        self.n_hop = n_hop
        self.padding = padding
        self.power_spectrogram = float(power_spectrogram)
        self.return_decibel_spectrogram = return_decibel_spectrogram
        super(Spectrogram, self).__init__(**kwargs)

    def build(self, input_shape):
        self.len_src = input_shape[1]
        self.n_ch = input_shape[2]
        self.is_mono = self.n_ch == 1
        if self.image_data_format == "channels_first":
            self.ch_axis_idx = 1
        else:
            self.ch_axis_idx = 3
        if self.len_src is not None:
            assert self.len_src >= self.n_dft, f"Hey! The input is too short! {self.len_src} >= {self.n_dft}"

        self.n_frame = conv_output_length(self.len_src, self.n_dft, self.padding, self.n_hop)

        dft_real_kernels, dft_imag_kernels = backend.get_stft_kernels(self.n_dft)
        self.dft_real_kernels = K.variable(dft_real_kernels, dtype=K.floatx(), name="real_kernels")
        self.dft_imag_kernels = K.variable(dft_imag_kernels, dtype=K.floatx(), name="imag_kernels")
        # kernels shapes: (filter_length, 1, input_dim, nb_filter)?
        if self.trainable_kernel:
            self.trainable_weights.append(self.dft_real_kernels)
            self.trainable_weights.append(self.dft_imag_kernels)
        else:
            self.non_trainable_weights.append(self.dft_real_kernels)
            self.non_trainable_weights.append(self.dft_imag_kernels)

        super(Spectrogram, self).build(input_shape)
        # self.built = True

    def compute_output_shape(self, input_shape):
        if self.image_data_format == "channels_first":
            return input_shape[0], self.n_ch, self.n_frame, self.n_filter
        else:
            return input_shape[0], self.n_frame, self.n_filter, self.n_ch

    def call(self, x):
        output = self._spectrogram_mono(x[:, :, 0:1])
        if self.is_mono is False:
            for ch_idx in range(1, self.n_ch):
                output = K.concatenate((output, self._spectrogram_mono(x[:, :, ch_idx : ch_idx + 1])), axis=self.ch_axis_idx)
        # output = output[..., 0]
        if self.power_spectrogram != 2.0:
            output = K.pow(K.sqrt(output), self.power_spectrogram)
        if self.return_decibel_spectrogram:
            output = backend_keras.amplitude_to_decibel(output)
        return output

    def get_config(self):
        config = {
            "n_dft": self.n_dft,
            "n_hop": self.n_hop,
            "padding": self.padding,
            "power_spectrogram": self.power_spectrogram,
            "return_decibel_spectrogram": self.return_decibel_spectrogram,
            "trainable_kernel": self.trainable_kernel,
            "image_data_format": self.image_data_format,
        }
        base_config = super(Spectrogram, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def _spectrogram_mono(self, x):
        """x.shape : (None, len_src, 1),
        returns 2D batch of a mono power-spectrogram"""
        x = K.expand_dims(x, 3)  # add a dummy dimension (channel axis)
        subsample = (self.n_hop, 1)
        output_real = K.conv2d(x, self.dft_real_kernels, strides=subsample, padding=self.padding, data_format="channels_last")
        output_imag = K.conv2d(x, self.dft_imag_kernels, strides=subsample, padding=self.padding, data_format="channels_last")
        output = output_real ** 2 + output_imag ** 2
        # now shape is (batch_sample, n_frame, 1, freq)
        if self.image_data_format == "channels_last":
            output = K.permute_dimensions(output, [0, 1, 3, 2])
        else:
            output = K.permute_dimensions(output, [0, 2, 3, 1])
        return output


class Melspectrogram(Spectrogram):
    """
        ### `Melspectrogram`
        ``python
        kapre.time_frequency.Melspectrogram(sr=22050, n_mels=128, fmin=0.0, fmax=None,
                                            power_melgram=1.0, return_decibel_melgram=False,
                                            trainable_fb=False, **kwargs)
        ``
    d
        Mel-spectrogram layer that outputs mel-spectrogram(s) in 2D image format.

        Its base class is `Spectrogram`.

        Mel-spectrogram is an efficient representation using the property of human
        auditory system -- by compressing frequency axis into mel-scale axis.

        #### Parameters
         * sr: integer > 0 [scalar]
           - sampling rate of the input audio signal.
           - Default: `22050`

         * n_mels: int > 0 [scalar]
           - The number of mel bands.
           - Default: `128`

         * fmin: float > 0 [scalar]
           - Minimum frequency to include in Mel-spectrogram.
           - Default: `0.0`

         * fmax: float > `fmin` [scalar]
           - Maximum frequency to include in Mel-spectrogram.
           - If `None`, it is inferred as `sr / 2`.
           - Default: `None`

         * power_melgram: float [scalar]
           - Power of `2.0` if power-spectrogram,
           - `1.0` if amplitude spectrogram.
           - Default: `1.0`

         * return_decibel_melgram: bool
           - Whether to return in decibel or not, i.e. returns log10(amplitude spectrogram) if `True`.
           - Recommended to use `True`, although it's not by default.
           - Default: `False`

         * trainable_fb: bool
           - Whether the spectrogram -> mel-spectrogram filterbanks are trainable.
           - If `True`, the frequency-to-mel matrix is initialised with mel frequencies but trainable.
           - If `False`, it is initialised and then frozen.
           - Default: `False`

         * htk: bool
           - Check out Librosa's `mel-spectrogram` or `mel` option.

         * norm: float [scalar]
           - Check out Librosa's `mel-spectrogram` or `mel` option.

         * **kwargs:
           - The keyword arguments of `Spectrogram` such as `n_dft`, `n_hop`,
           - `padding`, `trainable_kernel`, `image_data_format`.

        #### Notes
         * The input should be a 2D array, `(audio_channel, audio_length)`.
        E.g., `(1, 44100)` for mono signal, `(2, 44100)` for stereo signal.
         * It supports multichannel signal input, so `audio_channel` can be any positive integer.
         * The input shape is not related to keras `image_data_format()` config.

        #### Returns

        A Keras layer
         * abs(mel-spectrogram) in a shape of 2D data, i.e.,
         * `(None, n_channel, n_mels, n_time)` if `channels_first`,
         * `(None, n_mels, n_time, n_channel)` if `channels_last`,

    """

    def __init__(
        self,
        sr=22050,
        n_mels=128,
        fmin=0.0,
        fmax=None,
        power_melgram=1.0,
        return_decibel_melgram=False,
        trainable_fb=False,
        htk=False,
        norm=1,
        **kwargs,
    ):

        super(Melspectrogram, self).__init__(**kwargs)
        assert sr > 0
        assert fmin >= 0.0
        if fmax is None:
            fmax = float(sr) / 2
        assert fmax > fmin
        assert isinstance(return_decibel_melgram, bool)
        if "power_spectrogram" in kwargs:
            assert kwargs["power_spectrogram"] == 2.0, "In Melspectrogram, power_spectrogram should be set as 2.0."

        self.sr = int(sr)
        self.n_mels = n_mels
        self.fmin = fmin
        self.fmax = fmax
        self.return_decibel_melgram = return_decibel_melgram
        self.trainable_fb = trainable_fb
        self.power_melgram = power_melgram
        self.htk = htk
        self.norm = norm

    def build(self, input_shape):
        super(Melspectrogram, self).build(input_shape)
        self.built = False
        # compute freq2mel matrix -->
        mel_basis = backend.mel(
            self.sr, self.n_dft, self.n_mels, self.fmin, self.fmax, self.htk, self.norm
        )  # (128, 1025) (mel_bin, n_freq)
        mel_basis = np.transpose(mel_basis)

        self.freq2mel = K.variable(mel_basis, dtype=K.floatx())
        if self.trainable_fb:
            self.trainable_weights.append(self.freq2mel)
        else:
            self.non_trainable_weights.append(self.freq2mel)
        self.built = True

    def compute_output_shape(self, input_shape):
        if self.image_data_format == "channels_first":
            return input_shape[0], self.n_ch, self.n_mels, self.n_frame
        else:
            return input_shape[0], self.n_mels, self.n_frame, self.n_ch

    def call(self, x):
        power_spectrogram = super(Melspectrogram, self).call(x)
        # now,  channels_first: (batch_sample, n_ch, n_freq, n_time)
        #       channels_last: (batch_sample, n_freq, n_time, n_ch)
        print(x.shape, power_spectrogram.shape)
        if self.image_data_format == "channels_first":
            power_spectrogram = K.permute_dimensions(power_spectrogram, [0, 1, 2, 3])
        else:
            power_spectrogram = K.permute_dimensions(power_spectrogram, [0, 3, 1, 2])
        # now, whatever image_data_format, (batch_sample, n_ch, n_time, n_freq)
        output = K.dot(power_spectrogram, self.freq2mel)
        if self.image_data_format == "channels_first":
            output = K.permute_dimensions(output, [0, 1, 3, 2])
        else:
            output = K.permute_dimensions(output, [0, 3, 2, 1])
        if self.power_melgram != 2.0:
            output = K.pow(K.sqrt(output), self.power_melgram)
        if self.return_decibel_melgram:
            output = backend_keras.amplitude_to_decibel(output)
        return output

    def get_config(self):
        config = {
            "sr": self.sr,
            "n_mels": self.n_mels,
            "fmin": self.fmin,
            "fmax": self.fmax,
            "trainable_fb": self.trainable_fb,
            "power_melgram": self.power_melgram,
            "return_decibel_melgram": self.return_decibel_melgram,
            "htk": self.htk,
            "norm": self.norm,
        }
        base_config = super(Melspectrogram, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


def conv_output_length(input_length, filter_size, padding, stride, dilation=1):
    """Determines output length of a convolution given input length.
    # Arguments
        input_length: integer.
        filter_size: integer.
        padding: one of `"same"`, `"valid"`, `"full"`.
        stride: integer.
        dilation: dilation rate, integer.
    # Returns
        The output length (integer).
    """
    if input_length is None:
        return None
    assert padding in {"same", "valid", "full", "causal"}
    dilated_filter_size = (filter_size - 1) * dilation + 1
    if padding == "same":
        output_length = input_length
    elif padding == "valid":
        output_length = input_length - dilated_filter_size + 1
    elif padding == "causal":
        output_length = input_length
    elif padding == "full":
        output_length = input_length + dilated_filter_size - 1
    return (output_length + stride - 1) // stride



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#.
import os
import sys

sys.path.insert(0, os.path.abspath('../src'))

# -- Project information -----------------------------------------------------

project = 'DAS'
copyright = '2021, Jan Clemens'
author = 'Jan Clemens'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc", "sphinx.ext.coverage", "sphinx.ext.napoleon", "sphinx.ext.viewcode", "myst_nb",
    "sphinx.ext.autosummary", "sphinx_panels", "sphinxcontrib.images", 'sphinx.ext.autosectionlabel', "sphinx_inline_tabs"
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'README.md']

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = "sphinx"
# pygments_dark_style = "monokai"

html_logo = 'images/icon.png'
html_theme_options = {
    'display_version': False,
    'sidebar_hide_name': True,
    "collapse_navigation": False,
}


# Autosummary linkcode resolution
# https://www.sphinx-doc.org/en/master/usage/extensions/linkcode.html
def linkcode_resolve(domain, info):
    """Resolve GitHub URLs for linkcode extension."""

    if domain != "py":
        return None

    if not info["module"]:
        return None

    try:
        filename = docs.utils.resolve(info["module"], info["fullname"])
        if filename is None:
            return None
        return f"https://github.com/janclemenslab/das/blob/master/{filename}"
    except:
        print(info)
        raise


autosummary_generate = True
autoclass_content = 'both'

# Enable automatic role inference as a Python object for autodoc.
# This automatically converts object references to their appropriate role,
# making it much easier (and more legible) to insert references in docstrings.
#   Ex: `MyClass` -> :class:`MyClass`
# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-default_role
default_role = "py:obj"

myst_enable_extensions = [
    'colon_fence',
]

# jupyter_execute_notebooks = "off"
nb_execution_mode = "off"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'furo'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


"""Utils for loading and manipulating data for training and prediction."""
import numpy as np
import tensorflow.keras as keras
import dask.array
from typing import Optional, Callable, Sequence, List
from tqdm.autonotebook import tqdm


def unpack_batches(x: np.ndarray, padding: int = 0):
    """[summary]

    Args:
        x ([type]): [description]
        padding (int, optional): [description]. Defaults to 0.

    Returns:
        [type]: [description]
    """
    if padding > 0:
        x = x[:, padding:-padding, ...]
    # reshape everything from [batch, hist, classes] to [time, classes] <- ???
    x = x.reshape((-1, x.shape[-1]))
    return x


def get_data_from_gen(data_gen):
    # PREPEND data_padding samples of zeros/nans so we match len(x)? and postpend nans to match len(x) as well?
    x, y = data_gen.unroll(return_x=True, merge_batches=True)
    # reshape from [batches, nb_hist, ...] to [time, ...]
    x = unpack_batches(x, data_gen.data_padding)
    if y is not None:
        y = unpack_batches(y, data_gen.data_padding)
    return x, y


def sub_range(data_len, fraction: float, min_nb_samples: int = 0, seed=None):
    """[summary]

    Args:
        data_len (int): total length of data
        fraction (float): fraction of data_len to use
        seed (float): seed random number generator for reproducible subset selection

    Returns:
        first_sample (int), last_sample (int)
    """
    np.random.seed(seed)
    sub_len = int(max(np.ceil(fraction * data_len), np.ceil(min_nb_samples)))
    first_sample = np.random.randint(low=0, high=data_len - sub_len - 1)
    last_sample = first_sample + sub_len + 1
    return first_sample, last_sample


def compute_class_weights(y: np.ndarray) -> List[float]:
    """_summary_

    Args:
        y (np.ndarray): [T, nb_classes]

    Returns:
        np.ndarray: nb_classes
    """
    nb_classes = y.shape[1]

    # chunk y
    yy = dask.array.from_array(y)
    nb_chunks = len(yy.chunks[0])

    # count classes over blocks (dask.array.map_blocks does nto work fsr)
    counts = np.zeros((nb_chunks, nb_classes))
    for cnt, block in enumerate(tqdm(yy.blocks, total=nb_chunks, desc="Counting class occurrences")):
        counts[cnt, :] = np.sum(block.compute().astype(float), axis=0)

    # aggregate and normalize
    class_weights = np.sum(counts, axis=0)
    class_weights /= np.sum(class_weights)
    class_weights = [1 / class_weight for class_weight in class_weights]
    return class_weights


class AudioSequence(keras.utils.Sequence):
    """[summary]

    Methods:
        ...
    """

    def __init__(
        self,
        x: np.ndarray,
        y: Optional[np.ndarray] = None,
        batch_size: int = 32,
        shuffle: bool = True,
        nb_hist: int = 1,
        y_offset: Optional[int] = None,
        stride: int = 1,
        cut_trailing_dim: bool = False,
        with_y_hist: bool = False,
        data_padding: int = 0,
        first_sample: int = 0,
        last_sample: Optional[int] = None,
        output_stride: int = 1,
        nb_repeats: int = 1,
        shuffle_subset: Optional[float] = None,
        unpack_channels: bool = False,
        mask_input: Optional[int] = None,
        batch_processor: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        class_weights: Optional[Sequence[float]] = None,
        **kwargs
    ):
        """[summary]

        x and y can be mem-mapped numpy arrays or lazily loaded hdf5 (zarr, xarray) datasets. Dask arrays do not work since they are immutable.
        Args:
            x (np.ndarray): [nb_samples, ...]
            y (np.ndarray, optional): [nb_samples,  nb_classes] - class probabilities - so sum over classes for each sample should be 1.0. Defaults to None.
                                      If None, getitem will only return x batches - neither y nor sample weights
            batch_size (int, optional): number of batches to return. Defaults to 32.
            shuffle (bool, optional): randomize order of batches. Defaults to True.
            nb_hist (int, optional): nb of time steps per batch. Defaults to 1.
            y_offset (int, optional): time offset between x and y. nb_hist/2 if None (predict central sample in each batch). Defaults to None.
            stride (int, optional): nb of time steps between batches. Defaults to 1.
            cut_trailing_dim (bool, optional): Remove trailing dimension. Defaults to False.
            with_y_hist (bool, optional): y as central value of the x_hist window (False) or the full sequence covering
                                          the x_hist window (True). Defaults to False.
            data_padding (int, optional): if > 0, will set weight of as many samples at start and end of nb_hist window to zero. Defaults to 0.
            first_sample (int): 0
            last_sample (int, optional): None - last_sample in x, otherwise last_sample
            output_stride (int): Take every Nth sample as output. Useful in combination with a "downsampling frontend". Defaults to 1 (every sample).
            nb_repeats (int): Number of repeats before the dataset runs out of data. Defaults to 1 (no repeats).
            shuffle_subset (float): Fraction of batches to use - only works if shuffle=True
            unpack_channels (bool): For multi-channel models with single-channel preprocessing -
                                    unpack [nb_hist, nb_channels] -> [nb_channels * [nb_hist, 1]]
            mask_input (int): half width of the number of central samples to mask. Defaults to None (no masking).
            batch_processor (Callable[[np.ndarray], np.ndarray], optional): For augmentations. Defaults to None.
            class_weights (Sequence[float], optional): Weights for each class used for balancing. Defaults to None (no balancing).
        """
        # TODO clarify "channels" semantics
        self.x, self.y = x, y

        self.first_sample = first_sample
        self.last_sample = self.x.shape[0] if last_sample is None else last_sample
        self.nb_samples = self.last_sample - self.first_sample
        self.nb_repeats = nb_repeats
        self.output_stride = output_stride
        self.with_y = False if self.y is None else True
        if self.with_y:
            self.nb_classes = self.y.shape[-1]
        else:
            self.nb_classes = 0  # if no y data

        self.batch_size = batch_size
        self.stride = stride
        self.shuffle = shuffle
        self.shuffle_subset = shuffle_subset
        self.x_hist = nb_hist
        self.with_y_hist = with_y_hist
        self.data_padding = data_padding
        self.unpack_channels = unpack_channels
        self.class_weights = class_weights
        self.mask_input = mask_input
        s0 = self.first_sample / self.stride
        s1 = (self.last_sample - self.x_hist - 1) / self.stride
        self.allowed_batches = np.arange(s0, s1, dtype=np.uintp)  # choose from all existing batches
        if self.shuffle_subset is not None:  # only choose from a fixed subset of existing batches
            self.allowed_batches = np.random.choice(
                self.allowed_batches, size=int(len(self.allowed_batches) * self.shuffle_subset), replace=False
            )

        if y_offset is None:
            self.y_offset = int(self.x_hist / 2)
        else:
            self.y_offset = int(y_offset)

        # ignore padding samples at beginning and end of x_hist to avoid boundary conditions problems
        if self.with_y_hist:
            self.weights = np.ones((self.batch_size, self.x_hist))
            if self.data_padding > 0:
                self.weights[:, : self.data_padding] = 0
                self.weights[:, -self.data_padding :] = 0
            self.weights = self.weights[:, ::output_stride]
        else:
            self.weights = np.ones((self.batch_size,))

        self.batch_processor = batch_processor
        self._idx_offset = 0

    def unroll(self, return_x=True, merge_batches=True):
        """[summary]

        Args:
            return_x=True
            merge_batches=True

        Returns:
            [type]: [description]
            (xx, yy), (None, yy) or (xx,)
        """
        xx = None
        if return_x:
            xx = np.zeros((len(self), self.batch_size, self.x_hist, *self.x.shape[1:]))  # this is probably more general
        if self.with_y_hist:
            yy = (
                np.zeros((len(self), self.batch_size, int(self.x_hist / self.output_stride), self.nb_classes))
                if self.with_y
                else None
            )
        else:
            yy = np.zeros((len(self), self.batch_size, self.nb_classes)) if self.with_y else None

        for cnt, gen_output in enumerate(self):
            if return_x:
                if self.unpack_channels:
                    xx[cnt, ...] = np.concatenate(gen_output[0], axis=-1)
                else:
                    xx[cnt, ...] = gen_output[0]
            if self.with_y:
                yy[cnt, ...] = gen_output[1]

        if merge_batches:
            if return_x:
                xx = xx.reshape((len(self) * self.batch_size, self.x_hist, *self.x.shape[1:]))  # this is probably more general
            if self.with_y_hist:
                yy = (
                    yy.reshape((len(self) * self.batch_size, int(self.x_hist / self.output_stride), self.nb_classes))
                    if self.with_y
                    else None
                )
            else:
                yy = yy.reshape((len(self) * self.batch_size, self.nb_classes)) if self.with_y else None

        if self.with_y:
            out = (xx, yy)
        else:
            out = (xx,)
        return out

    def __len__(self):
        """Number of batches."""
        return int(
            self.nb_repeats
            * max(
                0,
                np.floor(
                    (self.nb_samples - ((self.stride * (self.batch_size - 1)) + self.x_hist))
                    / (self.stride * (self.batch_size))
                )
                + 1,
            )
        )

    def __str__(self):
        string = [
            "AudioSequence with {} batches each with {} items.\n".format(len(self), self.batch_size),
            "   Total of {} samples with\n".format(self.nb_samples),
            "   each x={} and\n".format(self.x.shape[1:]),
        ]
        string.append("   each y={}".format(self.y.shape[1:])) if self.y is not None else "no y."
        return "".join(string)

    def __getitem__(self, idx):
        """Get item from AudioSequence

        Args:
            idx (int): batch

        Returns:
            batch_x [np.ndarray]: [nb_batches, nb_hist, ..., nb_classes].
            batch_y [np.ndarray]: [nb_batches, nb_hist, nb_classes].
            weights_y [np.ndarray]: [nb_batches, nb_hist, nb_classes].
        """
        idx += self._idx_offset
        batch_x = np.zeros((self.batch_size, self.x_hist, *self.x.shape[1:]), dtype=self.x.dtype)

        if self.with_y:
            if self.with_y_hist:
                batch_y = np.zeros(
                    (self.batch_size, int(self.x_hist / self.output_stride), self.nb_classes), dtype=self.y.dtype
                )
            else:
                batch_y = np.zeros((self.batch_size, self.nb_classes), dtype=self.y.dtype)

        if self.shuffle:
            # pts = np.random.randint(self.first_sample / self.stride, (self.last_sample - self.x_hist - 1) / self.stride, self.batch_size)
            pts = np.random.choice(self.allowed_batches, size=self.batch_size, replace=False)
        else:
            pts = range(
                int(self.first_sample / self.stride) + idx * self.batch_size,
                int(self.first_sample / self.stride) + (idx + 1) * self.batch_size,
            )

        for cnt, bat in enumerate(pts):
            batch_x[cnt, ...] = self.x[int(bat * self.stride) : int(bat * self.stride + self.x_hist), ...].copy()

            if self.with_y:
                if self.with_y_hist:
                    batch_y[cnt, ...] = self.y[
                        int(bat * self.stride) : int(bat * self.stride + self.x_hist) : self.output_stride, ...
                    ]
                else:
                    batch_y[cnt, ...] = self.y[int(bat * self.stride + self.y_offset), ...]
        if self.unpack_channels:
            batch_x = [batch_x[..., chn][..., np.newaxis] for chn in range(batch_x.shape[-1])]

        # "mask" input
        if self.mask_input is not None:
            batch_x[:, int(batch_x.shape[1] / 2 - self.mask_input) : int(batch_x.shape[1] / 2 + self.mask_input), :] = 0

        if self.batch_processor is not None:
            batch_x = self.batch_processor(batch_x)

        if self.with_y:
            out = (batch_x, batch_y)

            if self.class_weights is not None:
                # weights contain weight for the class at each time point
                weights = np.zeros_like(self.weights)
                labels = np.argmax(batch_y, axis=-1)
                for label, weight in enumerate(self.class_weights):
                    weights[labels == label] = weight
                weights *= self.weights  # use weights as mask to zero boundaries
            else:
                weights = self.weights

            if self.data_padding > 0:
                out = (batch_x, batch_y, weights)
        else:
            out = (batch_x,)
        return out



from . import io, predict
import numpy as np
import sklearn.metrics
from tqdm.autonotebook import tqdm
import warnings
from joblib import Parallel, delayed
import itertools
from typing import List, Tuple, Dict, Union, Optional
import dask.array as da
import logging

logger = logging.getLogger(__name__)
# suppress UndefinedMetricWarnings in case of no preds
warnings.simplefilter("ignore")


def postprocess(class_probabilities, gap_dur=None, min_len=None, segment_dims=None, segment_names=None, segment_thres=None):
    probs_post = predict.predict_segments(
        class_probabilities,
        segment_fillgap=gap_dur,
        segment_minlen=min_len,
        segment_dims=segment_dims,
        segment_names=segment_names,
    )
    return probs_post["samples"]


def score(labels_true, labels_pred):
    return sklearn.metrics.f1_score(labels_true, labels_pred, average="macro")


def obj_fun(labels_true, probs_pred, gap_dur, min_len, segment_dims=None, segment_names=None):
    warnings.simplefilter("ignore")
    labels_pred = postprocess(probs_pred, gap_dur, min_len, segment_dims, segment_names)
    return score(labels_true, labels_pred)


def optimize_grid(
    labels_train_true, probs_train_pred, gap_durs, min_lens, segment_dims=None, segment_names=None, nb_workers=-1
):
    params = list(itertools.product(gap_durs, min_lens))
    f1_scores = Parallel(n_jobs=nb_workers)(
        delayed(obj_fun)(labels_train_true, probs_train_pred, gap_dur, min_len, segment_dims, segment_names)
        for gap_dur, min_len in tqdm(params, total=len(params))
    )
    # f1_scores = []
    # for gap_dur, min_len in tqdm(params, total=len(params)):
    #     val = obj_fun(labels_train_true, probs_train_pred, gap_dur, min_len, segment_dims, segment_names)
    #     f1_scores.append(val)
    best_idx = np.argmax(f1_scores)
    best_gap_dur, best_min_len, best_score = params[best_idx][0], params[best_idx][1], f1_scores[best_idx]
    return best_gap_dur, best_min_len, best_score, f1_scores


def optimize(
    dataset_path: str,
    model_save_name: str,
    gap_durs: List[float],
    min_lens: List[float],
    nb_workers: Optional[int] = -1,
) -> Tuple[float, float, Dict[str, Union[float, List[float]]]]:
    """[summary]

    Args:
        dataset_path (str): [description]
        model_save_name (str): [description]
        gap_durs (List[float], optional): in seconds.
        min_lens (List[float], optional): in seconds.
        nb_workers (int, optional): Number of parallel workers to use. Defaults to -1 (same as cores).

    Returns:
        Tuple[float, float, Dict[str, Union[float, List[float]]]]: [description]
    """

    data = io.npy_dir.load(dataset_path, memmap_dirs="all")
    fs = data.attrs["samplerate_x_Hz"]

    gap_durs = (gap_durs * fs).astype(int)
    min_lens = (min_lens * fs).astype(int)

    # get raw predictions to run post-processing on
    logger.info("   Generating raw predictions")
    _, segments, probs_train_pred, _ = predict.predict(data["train"]["x"], model_save_name=model_save_name, save_memory=True)
    labels_train_true = postprocess(
        da.from_array(data["train"]["y"]), segment_dims=segments["index"], segment_names=segments["names"]
    )

    ml = min(len(labels_train_true), len(probs_train_pred))

    # get the F1 score before post-processing
    labels_train_pred = postprocess(probs_train_pred, segment_dims=segments["index"], segment_names=segments["names"])
    score_train_pre = score(labels_train_true[:ml], labels_train_pred[:ml])

    # test all (gap_dur, min_len) combinations and choose the one maximizes the F1 score
    logger.info("   Grid search for optimal values")
    best_gap_dur, best_min_len, score_train, f1_scores = optimize_grid(
        labels_train_true[:ml],
        probs_train_pred[:ml],
        gap_durs=gap_durs,
        min_lens=min_lens,
        segment_dims=segments["index"],
        segment_names=segments["names"],
        nb_workers=nb_workers,
    )

    # validate found parameters on validation set
    if len(data["val"]["x"]):
        # get raw predictions to run post-processing on
        _, _, probs_val_pred, _ = predict.predict(data["val"]["x"], model_save_name=model_save_name, save_memory=True)
        labels_val_true = postprocess(
            da.from_array(data["val"]["y"]), segment_dims=segments["index"], segment_names=segments["names"]
        )

        ml = min(len(labels_val_true), len(probs_val_pred))

        # get the F1 score before post-processing
        labels_val_pred = postprocess(probs_val_pred, segment_dims=segments["index"], segment_names=segments["names"])
        score_val_pre = score(labels_val_true[:ml], labels_val_pred[:ml])

        # get score with best parameters
        score_val = obj_fun(
            labels_val_true[:ml],
            probs_val_pred[:ml],
            gap_dur=best_gap_dur,
            min_len=best_min_len,
            segment_dims=segments["index"],
            segment_names=segments["names"],
        )
    else:
        score_val_pre = None
        score_val = None

    # from samples to seconds
    best_gap_dur, best_min_len = float(best_gap_dur / fs), float(best_min_len / fs)

    scores = {
        "train_pre": float(score_train_pre),
        "train": float(score_train),
        "val_pre": float(score_val_pre),
        "val": float(score_val),
        "all_train": f1_scores,
    }

    return best_gap_dur, best_min_len, scores



import os


ignore_modules = [
    "das.cli",
    "dsa._models",
]

def make_api_doctree():
    doctree = ""

    for root, dirs, files in os.walk("../src/das"):
        # remove leading "../src/"
        root = root[7:]

        for file in sorted(files):
            if file.endswith(".py") and not file.startswith("_"):
                full = os.path.join(root, file)
                full = full[:-3].replace(os.sep, ".")

                ignore = False
                for ignore_module in ignore_modules:
                    if full.startswith(ignore_module):
                        ignore = True
                        break
                if not ignore:
                    doctree += f"   {full}\n"

    # get the api doc header
    with open("_templates/api_head.rst", "r") as f:
        api_head = f.read()

    # write file for api doc with header + doctree
    with open("api.rst", "w") as f:
        f.write("..\n  This file is auto-generated.\n\n")
        f.write(api_head)
        f.write(doctree)


if __name__ == "__main__":
    make_api_doctree()


import das.augmentation
import numpy as np
import das
import yaml


def test_param():
    das.augmentation.Constant(10)((3,2))
    das.augmentation.Normal(1, 2)((3,2))
    das.augmentation.Uniform(1, 2)((3,2))


def test_augmentation():
    signal = np.arange(0, 100) / 100
    signal = signal[np.newaxis, :, np.newaxis]

    aug_gain = das.augmentation.Gain(gain=das.augmentation.Uniform(-1, 1))
    aug_gain(signal)

    aug_offset = das.augmentation.Offset(offset=das.augmentation.Uniform(-1, 1))
    aug_offset(signal)

    aug_addnoise = das.augmentation.MaskNoise(mean=das.augmentation.Uniform(-1, 1),
                                            std=das.augmentation.Uniform(0, 0.1))
    aug_addnoise(signal)

    aug_masknoise = das.augmentation.MaskNoise(duration=das.augmentation.Uniform(5, 20),
                                            std=das.augmentation.Uniform(0, 0.1))
    aug_masknoise(signal)

    aug_masknoise = das.augmentation.MaskNoise(duration=das.augmentation.Uniform(5, 20),
                                            std=das.augmentation.Uniform(0, 0.1),
                                            add=False)
    aug_masknoise(signal)

    aug_horzflip = das.augmentation.HorizontalFlip(flip=das.augmentation.Uniform(-1, 1))
    aug_horzflip(signal)

    aug_upsamp = das.augmentation.Upsampling(factor=das.augmentation.Uniform(1, 1.5))
    aug_upsamp(signal)

    aug_circshift = das.augmentation.CircShift(shift=das.augmentation.Uniform(-10, 10))
    aug_circshift(signal)


def test_augmentations():
    aug_circshift = das.augmentation.CircShift(shift=das.augmentation.Uniform(-10, 10))
    aug_masknoise = das.augmentation.MaskNoise(duration=das.augmentation.Uniform(5, 20),
                                            std=das.augmentation.Uniform(0, 0.1))
    augs = [aug_circshift, aug_masknoise]
    aug_shiftnoise = das.augmentation.Augmentations(augs)

    signal = np.arange(0, 100) / 100
    signal = signal[np.newaxis, :, np.newaxis]
    aug_shiftnoise(signal)

def test_from_yaml():
    yaml_string ="""
    Gain:
      gain:
        Uniform:
          lower: 0.8
          upper: 1.25

    MaskNoise:
      std:
        Normal:
          mean: 0.1
          std: 1
        mean:
        Constant:
          value: 0

    HorizontalFlip:
      flip:
        Uniform:
        """
    aug_dict = yaml.safe_load(yaml_string)
    augs = das.augmentation.Augmentations.from_dict(aug_dict)

    signal = np.arange(0, 100) / 100
    signal = signal[np.newaxis, :, np.newaxis]
    augs(signal)
    assert len(augs) == 3



"""Utilities for interacting with DAS-Menagerie."""
from typing import Callable
import urllib.request
import tempfile
import numpy as np

# from github import Github
from typing import Optional


def _npz_loader(fname: str):
    out = {}
    with np.load(fname) as f:
        for key in f.keys():
            try:
                out[key] = f[key].astype(float)
            except ValueError:
                out[key] = f[key]
    return out


def load_data(url: str):
    with tempfile.NamedTemporaryFile() as fp:
        urllib.request.urlretrieve(url, fp.name)
        out = _npz_loader(fp.name)
    return out


# def download_model(name: str, to: str = '.'):
#     print(name, to)
#     url = name
#     urllib.request.urlretrieve(url, to)


# def list_models(gh_repo_url: str = "janclemenslab/das-menagerie", gh_access_token: Optional[str] = None):
#     G = Github(gh_access_token)
#     repo = G.get_repo(gh_repo_url)
#     releases = repo.get_releases()
#     trained_models = dict()
#     for release in releases:
#         name = release.tag_name
#         trained_models[name] = {}
#         trained_models[name]['zipball_url'] = release.zipball_url
#         trained_models[name]['models'] = []

#         # models: base_url, model_url, params_url
#         for asset in release.raw_data['assets']:
#             if asset['name'].endswith('_model.h5'):
#                 asset['basename'] = asset['name'][:-len('_model.h5')]
#                 # find accompanying params file
#                 for asset_param in release.raw_data['assets']:
#                     if asset_param['name'] == asset['basename'] + '_params.yaml':
#                         asset['params'] = asset_param
#                         asset['params_url'] = asset_param['url']
#                         asset['params_name'] = asset_param['name']
#                         asset['params_url_friendly'] = asset_param['browser_download_url']
#                 trained_models[name]['models'].append(asset)
#             elif asset['name'].endswith('_params.yaml'):
#                 asset_type = 'params'
#             elif asset['name'] == '/data.npz':
#                 asset_type = 'data'
#                 trained_models[name]['data'] = asset

#     return trained_models

# try:
#     trained_models = list_models()
# except Exception as e:
#     print(e)



"""Defines the network architectures."""
import tensorflow as tf
import tensorflow.keras as keras
import tensorflow.keras.layers as kl
from tensorflow.keras import regularizers
from tensorflow.keras.applications.resnet_v2 import ResNet50V2
from typing import List, Optional
from . import tcn as tcn_layer
from .kapre.time_frequency import Spectrogram
from . import spec_utils

model_dict = dict()


def _register_as_model(func):
    """Adds func to model_dict Dict[modelname: modelfunc]. For selecting models by string."""
    model_dict[func.__name__] = func
    return func


@_register_as_model
def tcn(*args, **kwargs):
    """Synonym for tcn_stft."""
    return tcn_stft(*args, **kwargs)


@_register_as_model
def tcn_stft(
    nb_freq: int,
    nb_classes: int,
    nb_hist: int = 1,
    nb_filters: int = 16,
    kernel_size: int = 3,
    nb_conv: int = 1,
    loss: str = "categorical_crossentropy",
    dilations: Optional[List[int]] = None,
    activation: str = "norm_relu",
    use_skip_connections: bool = True,
    return_sequences: bool = True,
    dropout_rate: float = 0.00,
    padding: str = "same",
    sample_weight_mode: str = None,
    nb_pre_conv: int = 0,
    pre_nb_dft: int = 64,
    nb_lstm_units: int = 0,
    learning_rate: float = 0.0005,
    upsample: bool = True,
    use_separable: bool = False,
    use_resnet: bool = False,
    compile: bool = True,
    **kwignored
):
    """Create TCN network with optional trainable STFT layer as pre-processing and downsampling frontend.

    Args:
        nb_freq (int): [description]
        nb_classes (int): [description]
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 1.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        dilations (List[int], optional): [description]. Defaults to [1, 2, 4, 8, 16].
        activation (str, optional): [description]. Defaults to 'norm_relu'.
        use_skip_connections (bool, optional): [description]. Defaults to True.
        return_sequences (bool, optional): [description]. Defaults to True.
        dropout_rate (float, optional): [description]. Defaults to 0.00.
        padding (str, optional): [description]. Defaults to 'same'.
        nb_pre_conv (int, optional): If >0 adds a single STFT layer with a hop size of 2**nb_pre_conv before the TCN.
                                     Useful for speeding up training by reducing the sample rate early in the network.
                                     Defaults to 0 (no downsampling)
        pre_nb_dft (int, optional): Duration of filters (in samples) for the STFT frontend.
                                    Number of filters is pre_nb_dft // 2 + 1.
                                    Defaults to 64.
        learning_rate (float, optional) Defaults to 0.0005
        nb_lstm_units (int, optional): Defaults to 0.
        upsample (bool, optional): whether or not to restore the model output to the input samplerate.
                                   Should generally be True during training and evaluation but may speed up inference.
                                   Defaults to True.
        use_separable (bool, optional): use separable convs in residual block. Defaults to False.
        use_resnet (bool, optional): Defaults to False.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred

    Returns:
        [keras.models.Model]: Compiled TCN network model.
    """
    if dilations is None:
        dilations = [1, 2, 4, 8, 16]

    input_layer = kl.Input(shape=(nb_hist, nb_freq))
    out = input_layer

    if nb_pre_conv > 0:
        out = Spectrogram(
            n_dft=pre_nb_dft,
            n_hop=2 ** nb_pre_conv,
            return_decibel_spectrogram=True,
            power_spectrogram=1.0,
            trainable_kernel=True,
            name="trainable_stft",
            image_data_format="channels_last",
        )(out)
        if not use_resnet:
            out = kl.Reshape((out.shape[1], out.shape[2] * out.shape[3]))(out)

    if use_resnet:
        out = kl.Activation("relu")(out)
        out = kl.Concatenate(axis=-1)([out, out, out])
        out = ResNet50V2(input_shape=out.shape[1:], weights="imagenet", include_top=False)(out)
        out = kl.Reshape((out.shape[1], out.shape[2] * out.shape[3]))(out)
        out = kl.BatchNormalization()(out)

    x = tcn_layer.TCN(
        nb_filters=nb_filters,
        kernel_size=kernel_size,
        nb_stacks=nb_conv,
        dilations=dilations,
        activation=activation,
        use_skip_connections=use_skip_connections,
        padding=padding,
        dropout_rate=dropout_rate,
        return_sequences=return_sequences,
        use_separable=use_separable,
    )(out)

    if nb_lstm_units > 0:
        x = kl.Bidirectional(kl.LSTM(units=nb_lstm_units, return_sequences=True))(x)

    x = kl.Dense(nb_classes, activation="softmax")(x)

    if nb_pre_conv > 0 and upsample:
        x = kl.UpSampling1D(size=2 ** nb_pre_conv)(x)

    output_layer = x

    model = keras.models.Model(input_layer, output_layer, name="TCN")
    if use_resnet:
        model.get_layer(name="trainable_stft").trainable = False
        model.get_layer(name="resnet50v2").trainable = False

    if compile:
        optimizer = keras.optimizers.Adam(learning_rate=learning_rate, clipnorm=1.0)
        model.compile(optimizer=optimizer, loss=loss, sample_weight_mode=sample_weight_mode)
    return model


@_register_as_model
def stft_res_dense(
    nb_freq: int,
    nb_classes: int,
    nb_hist: int = 1,
    sample_weight_mode: str = None,
    learning_rate: float = 0.0005,
    compile: bool = True,
    stft_compute: bool = False,
    resnet_compute: bool = False,
    resnet_train: bool = False,
    label_smoothing: float = 0,
    **kwignored
):
    """Create TCN network with optional trainable STFT layer as pre-processing and downsampling frontend.

    Args:
        nb_freq (int): [description]
        nb_classes (int): [description]
        nb_hist (int, optional): [description]. Defaults to 1.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        nb_pre_conv (int, optional): If >0 adds a single STFT layer with a hop size of 2**nb_pre_conv before the TCN.
                                     Useful for speeding up training by reducing the sample rate early in the network.
                                     Defaults to 0 (no downsampling)
        pre_nb_dft (int, optional): Duration of filters (in samples) for the STFT frontend.
                                    Number of filters is pre_nb_dft // 2 + 1.
                                    Defaults to 64.
        learning_rate (float, optional) Defaults to 0.0005
        stft_compute (bool, optional): Defaults to False.
        resnet_compute (bool, optional): Defaults to False.
        resnet_train (bool, optional): Fine tune resnet weights. Defaults to False.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred

    Returns:
        [keras.models.Model]: Compiled network model.
    """

    input_layer = kl.Input(shape=(nb_hist, nb_freq))
    out = input_layer

    # TODO compute STFT

    if resnet_compute:
        out = tf.stack((out, out, out), axis=-1)
        vision_model = ResNet50V2(input_shape=out.shape[1:], weights="imagenet", include_top=False)
        out = vision_model(out, training=False)
        out = kl.TimeDistributed(
            kl.Dense(min(32, 4 * nb_classes), activation="tanh", kernel_regularizer=regularizers.L1(1e-4))
        )(out)

    if len(out.shape) > 1:
        out = kl.Flatten()(out)

    out = kl.BatchNormalization()(out)
    # out = kl.Dropout(0.1)(out)
    out = kl.Dense(min(32, 4 * nb_classes), activation="tanh", kernel_regularizer=regularizers.L1(1e-4))(out)
    # out = kl.Dropout(0.1)(out)
    out = kl.Dense(2 * nb_classes, activation="tanh", kernel_regularizer=regularizers.L1(1e-4))(out)
    # out = kl.Dropout(0.1)(out)
    out = kl.Dense(nb_classes, activation="softmax")(out)

    output_layer = out

    model = keras.models.Model(input_layer, output_layer, name="RES")

    if resnet_compute and not resnet_train:
        model.get_layer("resnet50v2").trainable = False

    if compile:
        optimizer = keras.optimizers.Adam(learning_rate=learning_rate, clipnorm=1.0)
        model.compile(
            optimizer=optimizer,
            loss=keras.losses.CategoricalCrossentropy(label_smoothing=label_smoothing),
            sample_weight_mode=sample_weight_mode,
        )
    return model



"""Code for tuning the hyperparameters of a network."""
# TODO:
# write custom Tuner that generates datasets and overlap based nb_hist and kernel params
# see: https://keras.io/guides/keras_tuner/custom_tuner/
import time
import logging
import flammkuchen as fl
import numpy as np
from tensorflow.keras.callbacks import EarlyStopping, ReduceLROnPlateau, TensorBoard
from tensorflow import keras
import keras_tuner as kt
import yaml
import os
import sys
from typing import List, Optional, Tuple, Dict, Any
from . import data, models, utils, predict, io, evaluate, tracking, data_hash

logger = logging.getLogger(__name__)

try:  # fixes cuDNN error when using LSTM layer
    import tensorflow as tf

    physical_devices = tf.config.list_physical_devices("GPU")
    if physical_devices:
        tf.config.experimental.set_memory_growth(physical_devices[0], enable=True)
except Exception as e:
    logger.exception(e)


class TunableModel(kt.HyperModel):
    def __init__(self, params, tune_config=None):
        self.params = params.copy()
        self.tune_config = tune_config
        if self.tune_config is None:
            self.tune_config = {
                "nb_filters": [4, 8, 16, 32, 64],
                "kernel_size": [4, 8, 16, 32, 64],
                "learning_rate": [0.01, 0.001, 0.0001],
                "nb_hist": [128, 256, 512, 1024, 2048],
                "nb_conv": [1, 2, 3, 4, 6, 8],
            }

    def build(self, hp):
        for name, values in self.tune_config.items():
            hp.Choice(name, values=values)

        self.params.update(hp.values)
        model = models.model_dict[self.params["model_name"]](**self.params)
        return model


class DasTuner(kt.Tuner):
    def __init__(self, params, *args, tracker=None, **kwargs):
        self.params = params.copy()
        self.tracker = tracker
        super().__init__(*args, **kwargs)

    def run_trial(
        self,
        trial,
        train_x,
        train_y,
        val_x,
        val_y,
        epochs=10,
        steps_per_epoch=None,
        verbose=1,
        class_weight=None,
        callbacks=None,
    ):
        try:
            if callbacks is None:
                callbacks = []

            self.params.update(trial.hyperparameters.values)
            self.current_trial = trial

            # these need updating based on current hyperparameters
            self.params["stride"] = self.params["nb_hist"]
            if self.params["ignore_boundaries"]:
                # this does not completely avoid boundary effects but should minimize them sufficiently
                self.params["data_padding"] = int(np.ceil(self.params["kernel_size"] * self.params["nb_conv"]))
                self.params["stride"] = self.params["stride"] - 2 * self.params["data_padding"]

            data_gen = data.AudioSequence(
                train_x,
                train_y,
                shuffle=True,
                nb_repeats=1,
                last_sample=train_x.shape[0] - 2 * self.params["nb_hist"],
                **self.params,
            )
            val_gen = data.AudioSequence(val_x, val_y, shuffle=False, **self.params)
            logger.info("Data:")
            logger.info(f"training: {data_gen}")
            logger.info(f"validation: {val_gen}")

            logger.info("Hyperparameters:")
            logger.info(trial.hyperparameters.values)
            if steps_per_epoch is None:
                steps_per_epoch = min(len(data_gen), 1000)

            if self.tracker is not None:
                self.tracker.reinit(self.params)
            model = self.hypermodel.build(trial.hyperparameters)
            model.params = self.params  # attach params to model so we can save them in ModelParamsCheckpoint
            fit_hist = model.fit(
                data_gen,
                validation_data=val_gen,
                epochs=epochs,
                steps_per_epoch=steps_per_epoch,
                callbacks=callbacks,
                verbose=verbose,
            )
            return fit_hist
        except KeyboardInterrupt:
            logger.exception("Interrupted")
            sys.exit(0)
        except:
            logger.exception("Something went wrong. Will try to continue.")
            self.current_trial.status = "INVALID"


def train(
    *,
    data_dir: str,
    x_suffix: str = "",
    y_suffix: str = "",
    save_dir: str = "./",
    save_prefix: Optional[str] = None,
    save_name: Optional[str] = None,
    model_name: str = "tcn",
    nb_filters: int = 16,
    kernel_size: int = 16,
    nb_conv: int = 3,
    use_separable: List[bool] = False,
    nb_hist: int = 1024,
    ignore_boundaries: bool = True,
    batch_norm: bool = True,
    nb_pre_conv: int = 0,
    pre_nb_dft: int = 64,
    pre_kernel_size: int = 3,
    pre_nb_filters: int = 16,
    upsample: bool = True,
    dilations: Optional[List[int]] = None,
    nb_lstm_units: int = 0,
    verbose: int = 2,
    batch_size: int = 32,
    nb_epoch: int = 400,
    learning_rate: Optional[float] = None,
    reduce_lr: bool = False,
    reduce_lr_patience: int = 5,
    fraction_data: Optional[float] = None,
    seed: Optional[int] = None,
    batch_level_subsampling: bool = False,
    augmentations: Optional[str] = None,
    tensorboard: bool = False,
    wandb_api_token: Optional[str] = None,
    wandb_project: Optional[str] = None,
    wandb_entity: Optional[str] = None,
    log_messages: bool = False,
    nb_stacks: int = 2,
    with_y_hist: bool = True,
    balance: bool = False,
    version_data: bool = True,
    tune_config: Optional[str] = None,
    nb_tune_trials: int = 1_000,
    _qt_progress: bool = False,
) -> Tuple[keras.Model, Dict[str, Any]]:
    """Tune the hyperparameters of a DAS network.

    Args:
        data_dir (str): Path to the directory or file with the dataset for training.
                        Accepts npy-dirs (recommended), h5 files or zarr files.
                        See documentation for how the dataset should be organized.
        x_suffix (str): Select dataset used for training in the data_dir by suffix (`y_` + X_SUFFIX).
                        Defaults to '' (will use the standard data 'x')
        y_suffix (str): Select dataset used as a training target in the data_dir by suffix (`y_` + Y_SUFFIX).
                        Song-type specific targets can be created with a training dataset,
                        Defaults to '' (will use the standard target 'y')
        save_dir (str): Directory to save training outputs.
                        The path of output files will constructed from the SAVE_DIR, an optional SAVE_PREFIX,
                        and the time stamp of the start of training.
                        Defaults to the current directory ('./').
        save_prefix (Optional[str]): Prepend to timestamp.
                           Name of files created will be start with SAVE_DIR/SAVE_PREFIX + "_" + TIMESTAMP
                           or with SAVE_DIR/TIMESTAMP if SAVE_PREFIX is empty.
                           Defaults to '' (empty).
        save_name (Optional[str]): Append to prefix.
                           Name of files created will be start with SAVE_DIR/SAVE_PREFIX + "_" + SAVE_NAME
                           or with SAVE_DIR/SAVE_NAME if SAVE_PREFIX is empty.
                           Defaults to TIMESTAMP.
        model_name (str): Network architecture to use.
                          Use `tcn` (TCN) or `tcn_stft` (TCN with STFT frontend).
                          See das.models for a description of all models.
                          Defaults to `tcn`.
        nb_filters (int): Number of filters per layer.
                          Defaults to 16.
        kernel_size (int): Duration of the filters (=kernels) in samples.
                           Defaults to 16.
        nb_conv (int): Number of TCN blocks in the network.
                       Defaults to 3.
        use_separable (List[bool]): Specify which TCN blocks should use separable convolutions.
                                    Provide as a space-separated sequence of "False" or "True.
                                    For instance: "True False False" will set the first block in a
                                    three-block (as given by nb_conv) network to use separable convolutions.
                                    Defaults to False (no block uses separable convolutions).
        nb_hist (int): Number of samples processed at once by the network (a.k.a chunk duration).
                       Defaults to 1024 samples.
        ignore_boundaries (bool): Minimize edge effects by discarding predictions at the edges of chunks.
                                  Defaults to True.
        batch_norm (bool): Batch normalize.
                           Defaults to True.
        nb_pre_conv (int): Adds fronted with downsampling. The downsampling factor is `2**nb_pre_conv`.
                           The type of frontend depends on the model:
                           if model is `tcn`: adds a frontend of N conv blocks (conv-relu-batchnorm-maxpool2) to the TCN.
                           if model is `tcn_tcn`: adds a frontend of N TCN blocks to the TCN.
                           if model is `tcn_stft`: adds a trainable STFT frontend.
                           Defaults to 0 (no frontend, no downsampling).
        pre_nb_dft (int): Duration of filters (in samples) for the STFT frontend.
                          Number of filters is pre_nb_dft // 2 + 1.
                          Defaults to 64.
        pre_nb_filters (int): Number of filters per layer in the pre-processing TCN.
                              Defaults to 16.
        pre_kernel_size (int): Duration of filters (=kernels) in samples in the pre-processing TCN.
                               Defaults to 3.
        upsample (bool): whether or not to restore the model output to the input samplerate.
                         Should generally be True during training and evaluation but my speed up inference.
                         Defaults to True.
        dilations (List[int]): List of dilation rate, defaults to [1, 2, 4, 8, 16] (5 layer with 2x dilation per TCN block)
        nb_lstm_units (int): If >0, adds LSTM with `nb_lstm_units` LSTM units to the output of the stack of TCN blocks.
                             Defaults to 0 (no LSTM layer).
        verbose (int): Verbosity of training output (0 - no output during training, 1 - progress bar, 2 - one line per epoch).
                       Defaults to 2.
        batch_size (int): Batch size
                          Defaults to 32.
        nb_epoch (int): Maximal number of training epochs.
                        Training will stop early if validation loss did not decrease in the last 20 epochs.
                        Defaults to 400.
        learning_rate (Optional[float]): Learning rate of the model. Defaults should work in most cases.
                               Values typically range between 0.1 and 0.00001.
                               If None, uses model specific defaults: `tcn` 0.0001, `tcn_stft` and `tcn_tcn` 0.0005.
                               Defaults to None.
        reduce_lr (bool): Reduce learning rate when the validation loss plateaus.
                          Defaults to False.
        reduce_lr_patience (int): Number of epochs w/o a reduction in validation loss after which
                                  to trigger a reduction in learning rate.
                                  Defaults to 5 epochs.
        fraction_data (Optional[float]): Fraction of training and validation data to use.
                                         Defaults to 1.0.
        seed (Optional[int]): Random seed to reproducibly select fractions of the data.
                              Defaults to None (no seed).
        batch_level_subsampling (bool): Select fraction of data for training from random subset of shuffled batches.
                                        If False, select a continuous chunk of the recording.
                                        Defaults to False.
        tensorboard (bool): Write tensorboard logs to save_dir. Defaults to False.
        wandb_api_token (Optional[str]): API token for logging to wandb.
                                           Defaults to None (no logging to wandb).
        wandb_project (Optional[str]): Project to log to for wandb.
                                         Defaults to None (no logging to wandb).
        wandb_entity (Optional[str]): Entity (user or team) to log to for wandb.
                                        Defaults to None (no logging to wandb).
        log_messages (bool): Sets terminal logging level to INFO.
                             Defaults to False (will follow existing settings).
        nb_stacks (int): Unused if model name is `tcn`, `tcn_tcn`, or `tcn_stft`. Defaults to 2.
        with_y_hist (bool): Unused if model name is `tcn`, `tcn_tcn`, or `tcn_stft`. Defaults to True.
        balance (bool): Balance data. Weights class-wise errors by the inverse of the class frequencies.
                        Defaults to False.
        version_data (bool): Save MD5 hash of the data_dir to log and params.yaml.
                             Defaults to True (set to False for large datasets since it can be slow).
        tune_config (Optional[str]): Yaml file with key:value pairs defining the search space for tuning.
                                     Keys are parameter names, values are lists of possible parameter values.
        nb_tune_trials (int): Number of model variants to test during hyper parameter tuning. Defaults to 1_000.

        Returns
            model (keras.Model)
            params (Dict[str, Any])
    """
    if log_messages:
        logging.basicConfig(level=logging.INFO)

    # FIXME THIS IS NOT GREAT:
    sample_weight_mode = None
    data_padding = 0
    if with_y_hist:  # regression
        return_sequences = True
        stride = nb_hist
        y_offset = 0
        sample_weight_mode = "temporal"
        if ignore_boundaries:
            data_padding = int(
                np.ceil(kernel_size * nb_conv)
            )  # this does not completely avoid boundary effects but should minimize them sufficiently
            stride = stride - 2 * data_padding
    else:  # classification
        return_sequences = False
        stride = 1  # should take every sample, since sampling rates of both x and y are now the same
        y_offset = int(round(nb_hist / 2))

    if stride <= 0:
        raise ValueError(
            "Stride <=0 - needs to be >0. Possible solutions: reduce kernel_size, increase nb_hist parameters, uncheck ignore_boundaries"
        )

    if not upsample:
        output_stride = int(2 ** nb_pre_conv)
    else:
        output_stride = 1  # since we upsample output to original sampling rate. w/o upsampling: `output_stride = int(2**nb_pre_conv)` since each pre-conv layer does 2x max pooling

    if save_prefix is None:
        save_prefix = ""

    if len(save_prefix):
        save_prefix = save_prefix + "_"
    params = locals()
    del params["_qt_progress"]

    if tune_config is not None:
        with open(tune_config, "r") as stream:
            try:
                tune_config = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                logger.exception(exc)

    if stride <= 0:
        raise ValueError(
            "Stride <=0 - needs to be >0. Possible solutions: reduce kernel_size, increase nb_hist parameters, uncheck ignore_boundaries"
        )

    # remove learning rate param if not set so the value from the model def is used
    if params["learning_rate"] is None:
        del params["learning_rate"]

    if "_multi" in model_name:
        params["unpack_channels"] = True

    logger.info(f"Loading data from {data_dir}.")
    d = io.load(data_dir, x_suffix=x_suffix, y_suffix=y_suffix)
    # FIXME throw error if no `val` in `d``
    if "val" not in d:
        logging.error("No validation data found.")
        raise ValueError("No validation data found.")
    params.update(d.attrs)  # add metadata from data.attrs to params for saving

    if version_data:
        params["data_hash"] = data_hash.hash_data(data_dir)
        logger.info(f"Version of the data:")
        logger.info(f"   MD5 hash of {data_dir} is")
        logger.info(f"   {params['data_hash']}")

    if fraction_data is not None:
        if fraction_data > 1.0:  # seconds
            logger.info(
                f"{fraction_data} seconds corresponds to {fraction_data / (d['train']['x'].shape[0] / d.attrs['samplerate_x_Hz']):1.4f} of the training data."
            )
            fraction_data = np.min((fraction_data / (d["train"]["x"].shape[0] / d.attrs["samplerate_x_Hz"]), 1.0))
        elif fraction_data < 1.0:
            logger.info(f"Using {fraction_data:1.4f} of the training and validation data.")

    if fraction_data is not None and not batch_level_subsampling and fraction_data != 1.0:  # train on a subset
        min_nb_samples = nb_hist * (batch_size + 2)  # ensure the generator contains at least one full batch
        first_sample_train, last_sample_train = data.sub_range(
            d["train"]["x"].shape[0], fraction_data, min_nb_samples, seed=seed
        )
        first_sample_val, last_sample_val = data.sub_range(d["val"]["x"].shape[0], fraction_data, min_nb_samples, seed=seed)
    else:
        first_sample_train, last_sample_train = 0, None
        first_sample_val, last_sample_val = 0, None

    # TODO clarify nb_channels, nb_freq semantics - always [nb_samples,..., nb_channels] -  nb_freq is ill-defined for 2D data
    params.update(
        {
            "nb_freq": d["train"]["x"].shape[1],
            "nb_channels": d["train"]["x"].shape[-1],
            "nb_classes": len(params["class_names"]),
            "first_sample_train": first_sample_train,
            "last_sample_train": last_sample_train,
            "first_sample_val": first_sample_val,
            "last_sample_val": last_sample_val,
        }
    )
    logger.info("Parameters:")
    logger.info(params)

    logger.info("Preparing data")
    if fraction_data is not None and batch_level_subsampling:  # train on a subset
        np.random.seed(seed)
        shuffle_subset = fraction_data
    else:
        shuffle_subset = None

    params["class_weights"] = None
    if balance:
        logger.info("Balancing classes:")
        logger.info("   Computing class weights.")
        params["class_weights"] = data.compute_class_weights(d["train"]["y"][first_sample_train:last_sample_train])
        logger.info(f"   {params['class_weights']}")

    logger.info("Building network")

    os.makedirs(os.path.abspath(save_dir), exist_ok=True)
    if save_name is None:
        save_name = time.strftime("%Y%m%d_%H%M%S")
    save_name = "{0}/{1}{2}".format(save_dir, save_prefix, save_name)
    params["save_name"] = save_name

    logger.info(f"Will save to {save_name}.")

    # setup callbacks
    callbacks = [EarlyStopping(monitor="val_loss", patience=15)]

    if reduce_lr:
        callbacks.append(ReduceLROnPlateau(patience=reduce_lr_patience, verbose=1))

    if _qt_progress:
        callbacks.append(utils.QtProgressCallback(nb_epoch, _qt_progress))

    if tensorboard:
        callbacks.append(TensorBoard(log_dir=save_name))

    if wandb_api_token and wandb_project:  # could also get those from env vars!
        del params["wandb_api_token"]
        wandb = tracking.Wandb(wandb_project, wandb_api_token, wandb_entity, params)
        if wandb:
            callbacks.append(wandb.callback())
    else:
        wandb = None

    tuner = DasTuner(
        params=params,
        oracle=kt.oracles.BayesianOptimization(
            objective=kt.Objective("val_loss", "min"),
            max_trials=nb_tune_trials,
        ),
        hypermodel=TunableModel(params, tune_config),
        overwrite=False,
        directory=save_dir,
        project_name=os.path.basename(save_name),
        tracker=wandb,
    )

    logger.info(tuner.search_space_summary())
    utils.save_params(params, save_name)

    # TRAIN NETWORK
    logger.info("Start hyperparameter tuning")
    tuner.search(
        train_x=d["train"]["x"],
        train_y=d["train"]["y"],
        val_x=d["val"]["x"],
        val_y=d["val"]["y"],
        epochs=nb_epoch,
        verbose=verbose,
        callbacks=callbacks,
        class_weight=params["class_weights"],
    )
    logger.info(tuner.results_summary())

    # update params with best hp
    best_hp = tuner.get_best_hyperparameters()[0].values
    params.update(best_hp)

    logger.info(f"   Saving best params and model to {params['save_name']}.")
    utils.save_params(params, params["save_name"])
    model = models.model_dict[params["model_name"]](**params)
    model.save(params["save_name"] + "_model.h5")

    # TEST
    if "test" not in d or len(d["test"]["x"]) < nb_hist:
        logger.info("   No test data - skipping final evaluation step.")
    else:
        logger.info(f"   Re-loading last best model from {params['save_name']}.")

        logger.info("   Predicting")
        x_test, y_test, y_pred = evaluate.evaluate_probabilities(x=d["test"]["x"], y=d["test"]["y"], model=model, params=params)

        labels_test = predict.labels_from_probabilities(y_test)
        labels_pred = predict.labels_from_probabilities(y_pred)

        logger.info("   Evaluating")
        conf_mat, report = evaluate.evaluate_segments(labels_test, labels_pred, params["class_names"], report_as_dict=True)
        logger.info(conf_mat)
        logger.info(report)
        params["conf_mat"] = conf_mat.tolist()
        params["report"] = report
        logger.info(f'   Updating params file "{save_name}_params.yaml" with the test results.')
        utils.save_params(params, save_name)

        if wandb_api_token and wandb_project:  # could also get those from env vars!
            wandb.log_test_results(report)

        save_filename = "{0}_results.h5".format(save_name)
        logger.info("   Saving to " + save_filename + ".")
        ddd = {
            "fit_hist": [],
            "confusion_matrix": conf_mat,
            "classification_report": report,
            "x_test": x_test,
            "y_test": y_test,
            "y_pred": y_pred,
            "labels_test": labels_test,
            "labels_pred": np.array(labels_pred),
            "params": params,
        }
        fl.save(save_filename, ddd)

    logger.info("DONE.")
    return model, params, tuner



"""Segment (syllable) utilities."""
import numpy as np
from typing import Tuple


def fill_gaps(labels: np.ndarray, gap_dur: int = 100) -> np.ndarray:
    """Fill short gaps in a sequence of labelled samples.

    `---111111-1111---` -> `---11111111111---`

    Args:
        labels (np.ndarray): Sequence of labelled samples.
        gap_dur (int, optional): Minimal gap duration, in samples. Defaults to 100.

    Returns:
        np.ndarray: Labelled samples with short gaps filled.
    """
    onsets = np.where(np.diff(labels.astype(int)) == 1)[0]
    offsets = np.where(np.diff(labels.astype(int)) == -1)[0]
    if len(onsets) and len(offsets):
        onsets = onsets[onsets < offsets[-1]]
    if len(onsets) and len(offsets):
        offsets = offsets[offsets > onsets[0]]
    if len(onsets) and len(offsets) and len(onsets) == len(offsets):
        for idx in range(len(onsets)):
            if idx > 0 and offsets[idx - 1] > onsets[idx] - gap_dur:
                labels[offsets[idx - 1] : onsets[idx] + 1] = 1
    return labels


def remove_short(labels: np.ndarray, min_len: int = 100) -> np.ndarray:
    """Remove short syllables from sequence of labelled samples.

    `---1111-1---1--` -> `---1111--------`

    Args:
        labels (np.ndarray): Sequence of labelled samples.
        min_len (int, optional): Minimal segment (syllable) duration, in samples. Defaults to 100.

    Returns:
        np.ndarray: Labelled samples with short syllables removed.
    """
    onsets = np.where(np.diff(labels.astype(int)) == 1)[0]
    offsets = np.where(np.diff(labels.astype(int)) == -1)[0]
    if len(onsets) and len(offsets):
        onsets = onsets[onsets < offsets[-1]]
    if len(onsets) and len(offsets):
        offsets = offsets[offsets > onsets[0]]
    if len(onsets) and len(offsets) and len(onsets) == len(offsets):
        durations = offsets - onsets
        for onset, offset, duration in zip(onsets, offsets, durations):
            if duration < min_len:
                labels[onset : offset + 1] = 0
    return labels


def label_syllables_by_majority(
    labels: np.ndarray, onsets_seconds: np.ndarray, offsets_seconds: np.ndarray, samplerate: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Label syllables by a majority vote.

    Args:
        labels (np.ndarray): Sequence of dirty, per sample, labels.
        onsets_seconds (List[float]): Onset of each syllable in `labels`, in seconds.
        offsets_seconds (List[float]): Offset of each syllable in `labels`, in seconds.
        samplerate (float): Samplerate of `labels`, in Hz.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Sequence of syllables, clean sequence of per-sample labels.
    """
    syllables = []
    labels_clean = np.zeros_like(labels, dtype=int)

    onsets_sample = (onsets_seconds * samplerate).astype(int)
    offsets_sample = (offsets_seconds * samplerate).astype(int)

    for onset_sample, offset_sample in zip(onsets_sample, offsets_sample):
        # max sure the segment is at least one sample long
        os = max(onset_sample + 1, offset_sample)
        values, counts = np.unique(labels[onset_sample:os], return_counts=True)
        if len(values):
            majority_label = values[counts.argmax()]
            syllables.append(int(majority_label))
            labels_clean[onset_sample:offset_sample] = syllables[-1]

    syllables = np.array(syllables)

    return syllables, labels_clean


def levenshtein(seq1: str, seq2: str) -> float:
    """Compute the Levenshtein edit distance between two strings.

    Corresponds to the minimal number of insertions, deletions, and subsitutions
    required to transform `seq1` into `seq2`.

    Args:
        seq1 (str)
        seq2 (str)

    Returns:
        float: The Levenshtein distance between seq1 and seq1.
    """
    oneago = None
    thisrow = list(range(1, len(seq2) + 1)) + [0]
    for x in range(len(seq1)):
        twoago, oneago, thisrow = oneago, thisrow, [0] * len(seq2) + [x + 1]
        for y in range(len(seq2)):
            delcost = oneago[y] + 1
            addcost = thisrow[y - 1] + 1
            subcost = oneago[y - 1] + (seq1[x] != seq2[y])
            thisrow[y] = min(delcost, addcost, subcost)
    return thisrow[len(seq2) - 1]


def syllable_error_rate(true: str, pred: str) -> float:
    """Compute the Levenshtein edit distance normalized by length of `true`.

    Args:
        true (str): Ground truth labels for a sequence of syllables.
        pred (str): Predicted labels for a sequence of syllables.

    Raises:
        TypeError: if either input is not a `str`

    Returns:
        float: Levenshtein distance normalized by length of `true`.
    """

    if type(true) != str or type(pred) != str:
        raise TypeError("Both `true` and `pred` must be of type `str`")

    return levenshtein(pred, true) / len(true)



import tensorflow.keras.backend as K
import tensorflow.keras.layers
import tensorflow.keras as keras
from tensorflow.keras import optimizers

# from tensorflow.keras.engine.topology import Layer  #only used for type annotations
from tensorflow.keras.layers import Activation, Lambda
from tensorflow.keras.layers import Conv1D, SpatialDropout1D, SeparableConv1D
from tensorflow.keras.layers import Conv1D, Dense, Layer
from tensorflow.keras import Input, Model

from typing import List, Tuple


def channel_normalization(x):
    # type: (Layer) -> Layer
    """Normalize a layer to the maximum activation

    This keeps a layers values between zero and one.
    It helps with relu's unbounded activation

    Args:
        x: The layer to normalize

    Returns:
        A maximal normalized layer
    """
    max_values = K.max(K.abs(x), 2, keepdims=True) + 1e-5
    out = x / max_values
    return out


def wave_net_activation(x: Layer) -> Layer:
    """This method defines the activation used for WaveNet

    described in https://deepmind.com/blog/wavenet-generative-model-raw-audio/

    Args:
        x: The layer we want to apply the activation to

    Returns:
        A new layer with the wavenet activation applied
    """
    tanh_out = Activation("tanh")(x)
    sigm_out = Activation("sigmoid")(x)
    return keras.layers.multiply([tanh_out, sigm_out])


def residual_block(
    x: Layer,
    s: int,
    i: int,
    activation: str,
    nb_filters: int,
    kernel_size: int,
    padding: str = "causal",
    use_separable: bool = False,
    dropout_rate: float = 0,
    name: str = "",
) -> Tuple[Layer, Layer]:
    """Defines the residual block for the WaveNet TCN

    Args:
        x: The previous layer in the model
        s: The stack index i.e. which stack in the overall TCN
        i: The dilation power of 2 we are using for this residual block
        activation: The name of the type of activation to use
        nb_filters: The number of convolutional filters to use in this block
        kernel_size: The size of the convolutional kernel
        padding: The padding used in the convolutional layers, 'same' or 'causal'.
        use_separable: Use separable convolution
        dropout_rate: Float between 0 and 1. Fraction of the input units to drop.
        name: Name of the model. Useful when having multiple TCN.

    Returns:
        A tuple where the first element is the residual model layer, and the second
        is the skip connection.
    """

    original_x = x

    if use_separable:
        conv = SeparableConv1D(
            filters=nb_filters, kernel_size=kernel_size, dilation_rate=i, depth_multiplier=4, padding=padding
        )(x)
    else:
        conv = Conv1D(filters=nb_filters, kernel_size=kernel_size, dilation_rate=i, padding=padding)(x)

    if activation == "norm_relu":
        x = Activation("relu")(conv)
        x = Lambda(channel_normalization)(x)
    elif activation == "wavenet":
        x = wave_net_activation(conv)
    else:
        x = Activation(activation)(conv)

    x = SpatialDropout1D(dropout_rate)(x)

    # 1x1 conv.
    x = Conv1D(nb_filters, 1, padding="same")(x)
    res_x = keras.layers.add([original_x, x])
    return res_x, x


def process_dilations(dilations):
    def is_power_of_two(num):
        return num != 0 and ((num & (num - 1)) == 0)

    if all([is_power_of_two(i) for i in dilations]):
        return dilations

    else:
        new_dilations = [2 ** i for i in dilations]
        return new_dilations


class TCN:
    """Creates a TCN layer.

    Args:
        input_layer: A tensor of shape (batch_size, timesteps, input_dim).
        nb_filters: The number of filters to use in the convolutional layers.
        kernel_size: The size of the kernel to use in each convolutional layer.
        dilations: The list of the dilations. Example is: [1, 2, 4, 8, 16, 32, 64].
        nb_stacks : The number of stacks of residual blocks to use.
        activation: The activations to use (norm_relu, wavenet, relu...).
        use_skip_connections: Boolean. If we want to add skip connections from input to each residual block.
        use_separable: Boolean. Use separable convolutions in each residual block.
        return_sequences: Boolean. Whether to return the last output in the output sequence, or the full sequence.
        padding: The padding to use in the convolutional layers, 'causal' or 'same'.
        dropout_rate: Float between 0 and 1. Fraction of the input units to drop.
        name: Name of the model. Useful when having multiple TCN.

    Returns:
        A TCN layer.
    """

    def __init__(
        self,
        nb_filters=64,
        kernel_size=2,
        nb_stacks=1,
        dilations=None,
        activation="norm_relu",
        use_skip_connections=True,
        use_separable=False,
        padding="causal",
        dropout_rate=0.0,
        return_sequences=True,
        name="tcn",
    ):
        self.name = name
        self.return_sequences = return_sequences
        self.dropout_rate = dropout_rate
        self.use_skip_connections = use_skip_connections

        try:
            len(use_separable)
        except TypeError:
            use_separable = [use_separable] * nb_stacks

        while len(use_separable) < nb_stacks:
            use_separable.append(use_separable[-1])
        self.use_separable = use_separable

        self.activation = activation
        self.dilations = dilations
        self.nb_stacks = nb_stacks
        self.kernel_size = kernel_size
        self.nb_filters = nb_filters
        self.padding = padding

    def __call__(self, inputs):
        if self.dilations is None:
            self.dilations = [1, 2, 4, 8, 16, 32]
        x = inputs
        x = Conv1D(self.nb_filters, 1, padding=self.padding)(x)
        skip_connections = []
        for s in range(self.nb_stacks):
            for i in self.dilations:
                x, skip_out = residual_block(
                    x,
                    s,
                    i,
                    self.activation,
                    self.nb_filters,
                    self.kernel_size,
                    self.padding,
                    self.use_separable[s],
                    self.dropout_rate,
                )
                skip_connections.append(skip_out)
        if self.use_skip_connections:
            x = keras.layers.add(skip_connections)
        x = Activation("relu")(x)

        if not self.return_sequences:
            output_slice_index = -1
            x = Lambda(lambda tt: tt[:, output_slice_index, :])(x)
        return x



"""Plot utilities."""
import matplotlib.pyplot as plt
from matplotlib_scalebar.scalebar import ScaleBar
import matplotlib_scalebar
import matplotlib as mpl
import matplotlib.legend as mlegend
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_pdf import PdfPages
import colorcet
import numpy as np
import itertools
from typing import Optional, List


def scalebar(length, dx=1, units="", label=None, axis=None, location="lower right", frameon=False, **kwargs):
    """Add scalebar to axis.

    Usage:
        plt.subplot(122)
        plt.plot([0,1,2,3], [1,4,2,5])
        scalebar(0.5, 'femtoseconds', label='duration', location='lower right’)

    Args:
        length (float): Length of the scalebar in units of axis ticks - length of 1.0 corresponds to spacing between to major x-ticks
        dx (int, optional): Scale factor for length. E.g. if scale factor is 10, the scalebar of length 1.0 will span 10 ticks. Defaults to 1.
        units (str, optional): Unit label (e.g. 'milliseconds'). Defaults to ''.
        label (str, optional): Title for scale bar (e.g. 'Duration'). Defaults to None.
        axis (matplotlib.axes.Axes, optional): Axes to add scalebar to. Defaults to None (currently active axis - plt.gca()).
        location (str, optional): Where in the axes to put the scalebar (upper/lower/'', left/right/center). Defaults to 'lower right'.
        frameon (bool, optional): Add background (True) or not (False). Defaults to False.
        kwargs: location=None, pad=None, border_pad=None, sep=None,
                frameon=None, color=None, box_color=None, box_alpha=None,
                scale_loc=None, label_loc=None, font_properties=None,
                label_formatter=None, animated=False):

    Returns:
        Handle to scalebar object
    """

    if axis is None:
        axis = plt.gca()

    if "dimension" not in kwargs:
        kwargs["dimension"] = matplotlib_scalebar.dimension._Dimension(units)

    scalebar = ScaleBar(dx=dx, units=units, label=label, fixed_value=length, location=location, frameon=frameon, **kwargs)
    axis.add_artist(scalebar)
    return scalebar


def remove_axes(axis=None, all=False, which="tblr"):
    """Remove top & left border around plot or all axes & ticks.

    Args:
        axis (matplotlib.axes.Axes, optional): Axes to modify. Defaults to None (currently active axis - plt.gca()).
        all (bool, optional): Remove all axes & ticks (True) or top & left border only (False). Defaults to False.
    """
    if axis is None:
        axis = plt.gca()

    # Hide the right and top spines

    axis.spines["right"].set_visible(False)
    axis.spines["top"].set_visible(False)
    # Only show ticks on the left and bottom spines
    axis.yaxis.set_ticks_position("left")
    axis.xaxis.set_ticks_position("bottom")

    if all:
        # Hide the left and bottom spines
        axis.spines["left"].set_visible(False)
        axis.spines["bottom"].set_visible(False)
        # Remove all tick labels
        axis.yaxis.set_ticks([])
        axis.xaxis.set_ticks([])


def despine(which="tr", axis=None):

    sides = {"t": "top", "b": "bottom", "l": "left", "r": "right"}

    if axis is None:
        axis = plt.gca()

    # Hide the spines
    for side in which:
        axis.spines[sides[side]].set_visible(False)

    # Hide the tick marks and labels
    if "r" in which:
        axis.yaxis.set_ticks_position("left")

    if "t" in which:
        axis.xaxis.set_ticks_position("bottom")

    if "l" in which:
        axis.yaxis.set_ticks([])

    if "b" in which:
        axis.xaxis.set_ticks([])


import string
from itertools import cycle


def label_axes(fig=None, labels=None, loc=None, **kwargs):
    """
    Walks through axes and labels each.

    kwargs are collected and passed to `annotate`

    Parameters
    ----------
    fig : Figure
         Figure object to work on

    labels : iterable or None
        iterable of strings to use to label the axes.
        If None, lower case letters are used.

    loc : len=2 tuple of floats
        Where to put the label in axes-fraction units
    """
    if fig is None:
        fig = plt.gcf()

    if labels is None:
        labels = string.ascii_uppercase

    if "size" not in kwargs:
        kwargs["size"] = 13

    # if 'weight' not in kwargs:
    #     kwargs['weight'] = 'bold'
    # re-use labels rather than stop labeling
    labels = cycle(labels)
    if loc is None:
        loc = (-0.2, 0.9)
    for ax, lab in zip(fig.axes, labels):
        ax.annotate(lab, xy=loc, xycoords="axes fraction", **kwargs)


def downsample_plot(x, y, ds=20):
    """Reduces complexity of exported pdfs w/o impairing visual appearance.

    Modified from pyqtgraph - downsampleMethod=peak.
    Keeps peaks so the envelope of the waveform is preserved.
    """
    n = len(x) // ds
    x1 = np.empty((n, 2))
    x1[:] = x[: n * ds : ds, np.newaxis]
    x0 = x1.reshape(n * 2)
    y1 = np.empty((n, 2))
    y2 = y[: n * ds].reshape((n, ds))
    y1[:, 0] = y2.max(axis=1)
    y1[:, 1] = y2.min(axis=1)
    y0 = y1.reshape(n * 2)
    return x0, y0


class Pdf:
    """Thin wrapper around Autosaving variant"""

    def __init__(self, savename, autosave=True, style=None, **savefig_kws):
        """[summary]

        Args:
            savename ([type]): [description]
            autosave (bool, optional): [description]. Defaults to True.
            style ([type], optional): [description]. Defaults to None.
        """
        self.pdf = PdfPages(savename)
        self.autosave = autosave
        self.style = style
        self.orig = None
        self.savefig_kws = savefig_kws

    def __enter__(self):
        if self.style is not None:
            self.orig = mpl.rcParams.copy()
            mpl.style.use(self.style)
        self.pdf.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.autosave:
            self.pdf.savefig(**self.savefig_kws)
        if self.orig is not None:
            dict.update(mpl.rcParams, self.orig)
        self.pdf.__exit__(exc_type, exc_val, exc_tb)


# from https://stackoverflow.com/a/60345118/2301098
def tablelegend(ax, col_labels=None, row_labels=None, title_label="", *args, **kwargs):
    """
    Place a table legend on the axes.

    Creates a legend where the labels are not directly placed with the artists,
    but are used as row and column headers, looking like this:
    """
    # ```
    # title_label   | col_labels[1] | col_labels[2] | col_labels[3]
    # -------------------------------------------------------------
    # row_labels[1] |
    # row_labels[2] |              <artists go there>
    # row_labels[3] |
    # ```

    # Parameters
    # ----------

    # ax : `matplotlib.axes.Axes`
    #     The artist that contains the legend table, i.e. current axes instant.

    # col_labels : list of str, optional
    #     A list of labels to be used as column headers in the legend table.
    #     `len(col_labels)` needs to match `ncol`.

    # row_labels : list of str, optional
    #     A list of labels to be used as row headers in the legend table.
    #     `len(row_labels)` needs to match `len(handles) // ncol`.

    # title_label : str, optional
    #     Label for the top left corner in the legend table.

    # ncol : int
    #     Number of columns.

    # Other Parameters
    # ----------------

    # Refer to `matplotlib.legend.Legend` for other parameters.

    # """
    #################### same as `matplotlib.axes.Axes.legend` #####################
    handles, labels, extra_args, kwargs = mlegend._parse_legend_args([ax], *args, **kwargs)
    if len(extra_args):
        raise TypeError("legend only accepts two non-keyword arguments")

    if col_labels is None and row_labels is None:
        ax.legend_ = mlegend.Legend(ax, handles, labels, **kwargs)
        ax.legend_._remove_method = ax._remove_legend
        return ax.legend_
    #################### modifications for table legend ############################
    else:
        ncol = kwargs.pop("ncol")
        handletextpad = kwargs.pop("handletextpad", 0 if col_labels is None else -2)
        title_label = [title_label]

        # blank rectangle handle
        extra = [Rectangle((0, 0), 1, 1, fc="w", fill=False, edgecolor="none", linewidth=0)]

        # empty label
        empty = [""]

        # number of rows infered from number of handles and desired number of columns
        nrow = len(handles) // ncol

        # organise the list of handles and labels for table construction
        if col_labels is None:
            assert nrow == len(row_labels), "nrow = len(handles) // ncol = %s, but should be equal to len(row_labels) = %s." % (
                nrow,
                len(row_labels),
            )
            leg_handles = extra * nrow
            leg_labels = row_labels
        elif row_labels is None:
            assert ncol == len(col_labels), "ncol = %s, but should be equal to len(col_labels) = %s." % (ncol, len(col_labels))
            leg_handles = []
            leg_labels = []
        else:
            assert nrow == len(row_labels), "nrow = len(handles) // ncol = %s, but should be equal to len(row_labels) = %s." % (
                nrow,
                len(row_labels),
            )
            assert ncol == len(col_labels), "ncol = %s, but should be equal to len(col_labels) = %s." % (ncol, len(col_labels))
            leg_handles = extra + extra * nrow
            leg_labels = title_label + row_labels
        for col in range(ncol):
            if col_labels is not None:
                leg_handles += extra
                leg_labels += [col_labels[col]]
            leg_handles += handles[col * nrow : (col + 1) * nrow]
            leg_labels += empty * nrow

        # Create legend
        ax.legend_ = mlegend.Legend(
            ax, leg_handles, leg_labels, ncol=ncol + int(row_labels is not None), handletextpad=handletextpad, **kwargs
        )
        ax.legend_._remove_method = ax._remove_legend
        return ax.legend_


def imshow_text(data, labels=None, ax=None, color_high="w", color_low="k", color_threshold=50, skip_zeros=False):
    """Text labels for individual cells of an imshow plot

    Args:
        data ([type]): Color values
        labels ([type], optional): Text labels. Defaults to None.
        ax ([type], optional): axis. Defaults to plt.gca().
        color_high (str, optional): [description]. Defaults to 'w'.
        color_low (str, optional): [description]. Defaults to 'k'.
        color_threshold (int, optional): [description]. Defaults to 50.
        skip_zeros (bool, optional): [description]. Defaults to False.
    """
    if ax is None:
        ax = plt.gca()

    if labels is None:
        labels = data

    for x, y in itertools.product(range(data.shape[0]), range(data.shape[1])):
        if labels[y, x] == 0:
            continue
        ax.text(
            x, y, f"{labels[y, x]:1.0f}", ha="center", va="center", c=color_high if data[y, x] > color_threshold else color_low
        )


def bar_text(ax=None, spacing=-20, to_int=True):
    """Add labels to the end of each bar in a bar chart.

    Arguments:
        ax (matplotlib.axes.Axes): The matplotlib object containing the axes
            of the plot to annotate. Defaults to current axis.
        spacing (int): The distance between the labels and the bars. Defaults to -20 (inside bar)
    """
    if ax is None:
        ax = plt.gca()

    # For each bar: Place a label
    for rect in ax.patches:
        # Get X and Y placement of label from rect.
        y_value = rect.get_height()
        x_value = rect.get_x() + rect.get_width() / 2

        # Number of points between bar and label. Change to your liking.
        space = spacing
        # Vertical alignment for positive values
        va = "bottom"

        # If value of bar is negative: Place label below bar
        if y_value < 0:
            # Invert space to place label below
            space *= -1
            # Vertically align label at top
            va = "top"

        # Use Y value as label and format number with one decimal place
        if to_int:
            label = int(y_value)
        else:
            label = float(y_value)

        # Create annotation
        ax.annotate(
            label,  # Use `label` as label
            (x_value, y_value),  # Place label at end of the bar
            xytext=(0, space),  # Vertically shift label by `space`
            textcoords="offset points",  # Interpret `xytext` as offset in points
            ha="center",  # Horizontally center label
            va=va,
            color="w",
        )  # Vertically align label differently for
        # positive and negative values.


def generate_colors(nb_colors: int = 1, start_color=None, start=0, step=1):
    """[summary]

    Args:
        nb_colors (int, optional): [description]. Defaults to 1.
        start_color ([type], optional): [description]. Defaults to None.
        start (int, optional): [description]. Defaults to 0.
        step (int, optional): [description]. Defaults to 1.

    Returns:
        [type]: [description]
    """
    cmap = colorcet.palette["glasbey_light"][start::step]
    cmap = list(cmap)[:nb_colors]
    if start_color is not None:
        cmap.insert(0, start_color)
    return cmap


def annotate_events(event_seconds, event_names=None, tmin: float = 0, tmax: float = np.inf, color: Optional[List] = None):
    """Plot events as vertical lines to plot.

    Args:
        event_seconds (List[float]): List of event times in seconds
        event_names (List[str], optional): List of event names. Defaults to None.
        tmin (float, optional): Start of the range of events annotated in seconds. Defaults to 0.
        tmax (float, optional): End of the range of events annotated in seconds. Defaults to np.inf.
        color (optional): Either a single valid matplotlib color (in that case all segment types will have that color)
                      or a list of matplotlib colors, one for each segment type.
                      Defaults to None, in which case a list of distinct colors
                      will be created automatically from colorcet's 'glasbey_light' palette.
    Raises:
        ValueError: If the number of colors does not match the number of unique event names.
    """
    unique_names = list(set(event_names))
    if color is None:  # generate a color for each event name
        color = generate_colors(len(unique_names))
    elif not isinstance(color, (list, tuple)):  # use given color for each event name
        color = [color for _ in unique_names]

    event_seen = {name: False for name in unique_names}
    for seconds, name in zip(event_seconds, event_names):
        if seconds > tmin and seconds < tmax:
            if not event_seen[name]:
                label = name
                event_seen[name] = True
            else:
                label = None
            plt.axvline(seconds, c=color[unique_names.index(name)], alpha=0.5, label=label)


def annotate_segments(onset_seconds, offset_seconds, segment_names=None, tmin: float = 0, tmax: float = np.inf, color=None):
    """Plot segments as vertical lines to plot.

    Args:
        onset_seconds ([type]): [description]
        offset_seconds ([type]): [description]
        segment_names ([type], optional): [description]. Defaults to None.
        tmin (float, optional): [description]. Defaults to 0.
        tmax (float, optional): [description]. Defaults to np.inf.
        color (optional): Either a single valid matplotlib color (in that case all segment types will have that color)
                      or a list of matplotlib colors, one for each segment type.
                      Defaults to None, in which case a list of distinct colors
                      will be created automatically from colorcet's 'glasbey_light' palette.
    Raises:
        ValueError: If the number of colors does not match the number of unique segment names.
    """
    unique_names = list(set(segment_names))
    if color is None:  # generate a color for each event name
        color = generate_colors(len(unique_names))
    elif not isinstance(color, (list, tuple)):  # use given color for each event name
        color = [color for _ in unique_names]

    if len(color) != len(unique_names):
        raise ValueError(f"Number of colors ({len(color)} does not match number of segment types ({len(unique_names)}).")

    segment_seen = {name: False for name in unique_names}
    for on, off, name in zip(onset_seconds, offset_seconds, segment_names):
        if on > tmin and off < tmax:
            if not segment_seen[name]:
                label = name
                segment_seen[name] = True
            else:
                label = None
            plt.axvspan(xmin=on, xmax=off, facecolor=color[unique_names.index(name)], alpha=0.25, label=label)



"""
Load data for training/testing.
See doc/data.md for a description of the data schema.
"""
import os.path
from . import npy_dir


def _select(data, x_suffix, y_suffix):
    for lvl in ["test", "val", "train"]:
        if lvl in data:
            if "y_" + y_suffix in data[lvl]:
                data[lvl]["y"] = data[lvl]["y_" + y_suffix]
                if "eventtimes_" + y_suffix in data[lvl]:
                    data[lvl]["eventtimes"] = data[lvl]["eventtimes_" + y_suffix]
            if "x_" + x_suffix in data[lvl]:
                data[lvl]["x"] = data[lvl]["x_" + x_suffix]
                if "eventtimes_" + x_suffix in data[lvl]:
                    data[lvl]["eventtimes"] = data[lvl]["eventtimes_" + x_suffix]

    if f"samplerate_x_{x_suffix}_Hz" in data.attrs:
        data.attrs["samplerate_x_Hz"] = data.attrs[f"samplerate_x_{x_suffix}_Hz"]

    if "class_names_" + y_suffix in data.attrs and "class_types_" + y_suffix in data.attrs:
        data.attrs["class_names"] = data.attrs["class_names_" + y_suffix]
        data.attrs["class_types"] = data.attrs["class_types_" + y_suffix]
    return data


def _to_dict(data):
    "Convert dict-like zarr or h5 store `data` to python dictionary."
    d = npy_dir.DictClass()
    d.attrs = dict(data.attrs)  # cast to dict since data.attrs are read-only for zarr stores
    for key_top in data.keys():
        d[key_top] = dict()
        for key, val in data[key_top].items():
            d[key_top][key] = val
    return d


def load(location, x_suffix="", y_suffix=""):
    """Load data for training/testing from zarr store, npy directory, or hdf5 file.

    Args:
        location ([type]): [description]
        x_suffix, y_suffix (str, optional): alternative key for the training source and target (allows for different x/y's for the same y/x in one data file)
    Returns:
        dict-like complying with data schema defined above
    """

    location = os.path.normpath(location)  # remove trailing path separators
    if location.endswith(".zarr"):
        import zarr

        data = zarr.open(location, mode="r")
    elif location.endswith(".h5"):
        import h5py

        data = h5py.File.open(location, mode="r")
    elif location.endswith(".npy"):
        data = npy_dir.load(location)
    else:
        raise ValueError(
            f'Could not load data. Location {location} has unknown extension - needs to end either in ".zarr", ".npy", or ".h5".'
        )

    data = _to_dict(data)

    if len(x_suffix) or len(y_suffix):
        data = _select(data, x_suffix, y_suffix)
    return data



"""Utilities for logging training runs.

We currenlty have integrations for `tensorboard <https://www.tensorflow.org/tensorboard>`_, and `wandb.ai <https://wandb.ai>`_.
While tensorboard is integrated with tensorflow. To use wandb you'll have to install the wandb API:
"pip install wandb" or "conda install wandb -c conda-forge".
"""
import logging
import os
from typing import Optional, Dict

try:
    import wandb
    from wandb.keras import WandbCallback

    HAS_WANDB = True
except ImportError as e:
    logging.debug("Could not import neptune libraries.")
    HAS_WANDB = False
HAS_WANDB = True


class Wandb:
    """Utility class for logging to wandb.ai during training."""

    def __init__(
        self,
        project: Optional[str] = None,
        api_token: Optional[str] = None,
        entity: Optional[str] = None,
        params: Optional[Dict] = None,
        infer_from_env: bool = False,
    ):
        """
        Args:
            project (Optional[str], optional): Project to log to. Defaults to None.
            api_token (Optional[str], optional):  api token. Defaults to None.
            entity (Optional[str], optional):  Entity (user/team name). Defaults to None.
            params (Optional[Dict], optional): Dict to log to `config`. Defaults to None.
            infer_from_env (bool, optional): read project and api_token from environment variables
                                             WANDB_PROJECT and WANDB_API_TOKEN.
                                             Defaults to False.
        """
        if not HAS_WANDB:
            self.run = None
            logging.error("Could not import wandb in das.tracking.")
            return

        try:
            if project is None:
                project = os.environ["WANDB_PROJECT"]
            if api_token is None:
                api_token = os.environ["WANDB_API_TOKEN"]

            wandb.login(key=api_token)
            self.project = project
            self.entity = entity
            self.run = wandb.init(project=self.project, entity=self.entity, settings=wandb.Settings(start_method="fork"))

            if params is not None:
                wandb.config.update(params)

        except:
            self.run = None
            logging.exception("Wandb stuff went wrong.")

    def reinit(self, params=None):
        self.run = wandb.init(reinit=True, project=self.project, entity=self.entity)
        if params is not None:
            wandb.config.update(params)

    def finish(self):
        self.run.finish()

    def callback(self, save_model=False):  # -> Optional[WandbCallback]:
        """Get callback for auto-logging from tensorfow/keras."""
        # CHECK: Is callback re-usable across reinits?
        if self.run is not None:
            return WandbCallback(save_model=save_model)
        else:
            pass

    def log_test_results(self, report: Dict):
        """Log final classification result from test data.

        Args:
            report (Dict): dictionary containing the classification report.
        """
        if self.run is not None:
            wandb.summary.update(report)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import numpy as np
from tensorflow.keras import backend as K


def amplitude_to_decibel(x, amin: float = 1e-10, dynamic_range: float = 80.0):
    """Convert (linear) amplitude to decibel (log10(x)).

        >>> x[x<amin] = amin  # clip everythin below amin
        >>> y = 10 * log(x) / log(10)  # log transform
        >>> y = ...  # rescale dyn range to [-80, 0]

    Args:
        x (Tensor): Tensor
        amin (float, optional): Minimal amplitude. Smaller values are clipped to this. Defaults to 1e-10 (dB).
        dynamic_range (float, optional): Dynamic range. Defaults to 80.0 (dB).

    Returns:
        [Tensor]: Tensor with the values converted to dB
    """
    log_spec = 10 * K.log(K.maximum(x, amin)) / np.log(10).astype(K.floatx())
    if K.ndim(x) > 1:
        axis = tuple(range(K.ndim(x))[1:])
    else:
        axis = None

    log_spec = log_spec - K.max(log_spec, axis=axis, keepdims=True)  # [-?, 0]
    log_spec = K.maximum(log_spec, -1 * dynamic_range)  # [-80, 0]
    return log_spec



"""Waveform level augmentations.

Individual implementations of `Augmentation` are callables that accept
the signal to augment as an argument and return the augmented signal.

Augmentation parameters can be `Constant`, or random with `Normal` or `Uniform` distribution.
Random parameters will be sampled from a given distribution anew for each augmentation.

`aug = Gain(gain=Normal(mean=1, std=0.5)); augmented_signal = aug(signal)`

Can be configured using a yaml file:
```yaml
Gain:  # Name of the augmentation class
  gain:  # arg for the augmentation class
    Uniform:  # Param type
       lower: 0.5  # param args
       upper: 2  # param args

MaskNoise:
  std:
    Normal:
      mean: 0
      std: 0.05
  mean:
    Constant:
       value: 0

NotchFilter:
  freq:  # Param-type arg
    Uniform:
      lower: 100
      upper: 600
  Q: 30  # standard arg
  samplerate_Hz: 10_000  # standard arg
```
Caution: You need to add a suffix starting with '-' (like "MaskNoise-1") to the class name
         if you want to use a class multiple times

`augs = Augmentations.from_yaml(filename)`
"""
import numpy as np
from typing import List, Optional
import scipy.signal
import yaml
from dataclasses import dataclass
from typing import Dict
import logging

logger = logging.getLogger(__name__)

aug_dict = dict()
params_dict = dict()


def _register_augmentation(func):
    """Adds func to model_dict Dict[augname: augfunc]. For selecting augs by string."""
    aug_dict[func.__name__] = func
    return func


def _register_param(func):
    """Adds func to model_dict Dict[paramname: paramfunc]. For selecting params by string."""
    params_dict[func.__name__] = func
    return func


@dataclass
class Param:
    """Base class for all parameters.

    Parameters are callables that return parameter values.
    """

    def __call__(self) -> np.ndarray:
        ...


@_register_param
class Constant(Param):
    """Constant parameter."""

    def __init__(self, value: float = 0.0):
        """
        Args:
            value (float, optional): Constant parameter value. Defaults to 0.0.
        """
        self.value = value

    def __call__(self, shape=1) -> np.ndarray:
        return self.value * np.ones(shape)

    def __str__(self):
        return f"{self.__class__.__name__}(value={self.value})"

    def __repr__(self):
        return f"{self.__class__.__name__}(value={self.value})"


@_register_param
class Normal(Param):
    """Normally distributed parameter."""

    def __init__(self, mean: float = 0.0, std: float = 1.0):
        """
        Args:
            mean (float, optional): Mean of the Normal pdf. Defaults to 0.0.
            std (float, optional): Standerd deviation of the Normal pdf. Defaults to 1.0.
        """
        self.mean = mean
        self.std = std

    def __call__(self, shape=1) -> np.ndarray:
        return np.random.normal(self.mean, self.std, size=shape)

    def __str__(self):
        return f"{self.__class__.__name__}(mean={self.mean}, std={self.std})"

    def __repr__(self):
        return f"{self.__class__.__name__}(mean={self.mean}, std={self.std})"


@_register_param
class Uniform(Param):
    """Uniformly distributed parameter."""

    def __init__(self, lower: float = -1.0, upper: float = 1.0):
        """
        Args:
            lower (float, optional): Lower bound. Defaults to -1.0.
            upper (float, optional): Upper bound. Defaults to 1.0.
        """
        self.lower = lower
        self.upper = upper

    def __call__(self, shape=1) -> np.ndarray:
        return np.random.uniform(self.lower, self.upper, size=shape)

    def __str__(self):
        return f"{self.__class__.__name__}(lower={self.lower}, upper={self.upper})"

    def __repr__(self):
        return f"{self.__class__.__name__}(lower={self.lower}, upper={self.upper})"


@dataclass
class Augmentation:
    """Base class for all augmentations.

    Augmentations are callables that return the augmented input.
    Can optionally pass through a second input.

    batch_x is expected to be [batch, time, channel].
    _apply works on [time, channel] inputs
    """

    def __call__(self, batch_x, batch_y=None):
        this_batch_x = batch_x.copy()
        for batch in range(batch_x.shape[0]):
            this_batch_x[batch, ...] = self._apply(this_batch_x[batch, ...])
        if batch_y is None:
            return this_batch_x
        else:
            return this_batch_x, batch_y

    def _apply(self, x: np.ndarray) -> np.ndarray:
        return x


@_register_augmentation
class Gain(Augmentation):
    """Multiply signal with gain factor."""

    def __init__(self, gain: Param):
        """
        Args:
            gain (Param): Gain.
        """
        self.gain = gain

    def _apply(self, x):
        x *= self.gain()
        return x

    # def __repr__(self):
    #     return f"{self.__class__.__name__}(gain={self.gain.__repr__()})"


@_register_augmentation
class NormalizeMax(Augmentation):
    """Multiply signal with gain factor."""

    def __init__(self):
        pass

    def _apply(self, x):
        x /= np.nanmax(np.abs(x), initial=1.0)
        return x


@_register_augmentation
class NormalizePercentile(Augmentation):
    """Multiply signal with gain factor."""

    def __init__(self, percentile: Param):
        self.percentile = percentile

    def _apply(self, x):
        x /= np.nanpercentile(x, self.percentile())
        return x


@_register_augmentation
class NormalizeStd(Augmentation):
    """Multiply signal with gain factor."""

    def __init__(self):
        pass

    def _apply(self, x):
        x /= np.nanstd(x)
        return x


@_register_augmentation
class Offset(Augmentation):
    """Add horizontal offset."""

    def __init__(self, offset: Param):
        """
        Args:
            offset (Param): Offset.
        """
        self.offset = offset

    def _apply(self, x):
        x = x + self.offset()
        return x


@_register_augmentation
class HorizontalFlip(Augmentation):
    """Horizontally flip signal."""

    def __init__(self, flip: Param):
        """
        Args:
            flip (Param): Signal is flipped if >0.
        """
        self.flip = flip

    def _apply(self, x):
        if self.flip() > 0:
            x *= -1
        return x


@_register_augmentation
class Upsampling(Augmentation):
    """Upsample signal."""

    def __init__(self, factor: Param):
        """
        Args:
            factor (Param): Upsampling factor.

        Raises:
            ValueError: if 'lower' attr of factor is <1 (would correspond to downsampling).
        """
        if hasattr(factor, "lower") and factor.lower < 1:  # type: ignore
            raise ValueError(f"Factor is {factor} - at the moment only upsampling (factors >= 1) is allowed.")
        self.factor = factor

    def _apply(self, x):
        len_x = x.shape[0]
        x = scipy.signal.resample_poly(x, up=int(len_x * self.factor()), down=len_x)
        x = x[:len_x]
        return x


@_register_augmentation
class MaskNoise(Augmentation):
    """Add noise or replace signal by noise for the full duration or a part of it."""

    def __init__(
        self, std: Optional[Param] = None, mean: Optional[Param] = None, duration: Optional[Param] = None, add: bool = True
    ):
        """
        Args:
            std (Optional[Param]): std of noise. Defaults to 1.
            mean (Optional[Param]): mean of noise. Defaults to 0.
            duration (Optional[Param]): nb_samples, Optional. If omitted will mask full duration.
            add (bool): add or replace. Defaults to True.
        """
        if std is None:
            std = Constant(1)
        self.std = std

        if mean is None:
            mean = Constant(0)
        self.mean = mean

        self.duration = duration
        self.add = add

    def _apply(self, x):
        len_x = x.shape[0]
        if self.duration is None:
            mask_start = 0
            duration = len_x
        else:
            duration = int(self.duration())
            mask_start = np.random.randint(low=0, high=len_x - duration)
        noise = np.random.randn(duration, *x.shape[1:]) * self.std() + self.mean()
        if self.add:
            x[mask_start : mask_start + duration, :] += noise
        else:
            x[mask_start : mask_start + duration, :] = noise
        return x

    def __repr__(self):
        return f"{self.__class__.__name__}(std={self.mean}, mean={self.mean})"


@_register_augmentation
class MaskMean(Augmentation):
    """Replaces stretch of `duration` samples with mean over that stretch."""

    def __init__(self, duration: Param):
        """
        Args:
            duration (Param): Duration of the stretch, in samples.
        """
        self.duration = duration

    def _apply(self, x):
        len_x = x.shape[0]
        duration = int(self.duration())
        mask_start = np.random.randint(low=0, high=len_x - duration)
        x[mask_start : mask_start + duration, :] = np.mean(x[mask_start : mask_start + duration], axis=0)
        return x


@_register_augmentation
class CircShift(Augmentation):
    """Circularly shift input along the first axis."""

    def __init__(self, shift: Param):
        """
        Args:
            shift (Param): Amount of shift, in samples.
        """
        self.shift = shift

    def _apply(self, x):
        x = np.roll(x, self.shift().astype(int), axis=0)
        return x


@_register_augmentation
class NotchFilter(Augmentation):
    """Notch filter."""

    def __init__(self, freq: Param, Q: float = 30, samplerate_Hz: float = 10_000):
        """
        Args:
            freq (Param): freq to remove, in Hz.
        """
        self.freq = freq
        self.Q = Q
        self.samplerate_Hz = samplerate_Hz

    def _apply(self, x):
        b, a = scipy.signal.iirnotch(self.freq(), self.Q, self.samplerate_Hz)
        x = scipy.signal.filtfilt(b, a, x, axis=0)
        return x


class Augmentations:
    """Bundles several augmentations."""

    def __init__(self, augmentations: List[Augmentation]):
        """
        Args:
            augmentations (List[Augmentation]): List of Augmentation instances.
        """
        self.augmentations = augmentations

    def __call__(self, batch_x, batch_y=None):
        for augmentation in self.augmentations:
            batch_x = augmentation(batch_x)
        if batch_y is None:
            return batch_x
        else:
            return batch_x, batch_y

    def __len__(self):
        return len(self.augmentations)

    @classmethod
    def from_yaml(cls, filename: str):
        aug_spec = yaml.safe_load(open(filename, "r"))
        return cls.from_dict(aug_spec)

    @classmethod
    def from_dict(cls, aug_spec: Dict):
        augs = []
        for name, args in aug_spec.items():
            name = name.split("-", 1)[0]  # split off "-WHATEVER" suffix
            params = dict()
            if args is not None:  # for augs without args
                for a_name, a_arg in args.items():
                    if isinstance(a_arg, dict):  # Param-type arg
                        p_name = list(a_arg.keys())[0]
                        p_args = a_arg[p_name]
                        # if no args for Params are provided, use defaults
                        if p_args is None:
                            p_args = {}
                        params[a_name] = params_dict[p_name](**p_args)
                    else:  # standard arg
                        params[a_name] = a_arg

            logger.debug(params)
            aug = aug_dict[name](**params)
            logger.debug(aug)
            augs.append(aug)
        return cls(augs)



"""Utilities for handling pulses."""
import numpy as np
import scipy.signal as ss
from typing import List, Tuple


def normalize_pulse(pulse: np.ndarray, smooth_win: int = 15, flip_win: int = 10) -> np.ndarray:
    """Normalize pulses.

    1. scales to unit-norm,
    2. aligns to energy maximum,
    3. flips so that pre-peak mean is positive

    Args:
        pulse (np.ndarray): should be [T,]
        smooth_win (int, optional): n samples of rect window used to smooth squared pulse for peak detection
        flip_win (int, optional): number of samples pre-peak used for determining sign of pulse for flipping.

    Returns:
        np.ndarray: normalized pulse [T,]
    """
    # scale
    pulse /= np.linalg.norm(pulse)
    pulse_len = len(pulse)
    pulse_len_half = int(pulse_len / 2)
    # center
    gwin = ss.windows.boxcar(int(smooth_win))
    pulse_env = np.convolve(pulse ** 2, gwin, mode="valid")
    offset = np.argmax(pulse_env) + int(np.ceil(smooth_win / 2)) + 1
    pulse = np.pad(pulse, (len(pulse) - offset, offset), mode="constant", constant_values=0)
    # flip
    if np.sum(pulse[pulse_len - flip_win : pulse_len]) < 0:
        pulse *= -1
    return pulse[pulse_len_half:-pulse_len_half]


def center_of_mass(x: np.ndarray, y: np.ndarray, thres: float = 0.5) -> float:
    """Calculate center of mass of y.

    Args:
        x (np.ndarray):
        y (np.ndarray):
        thres (float, optional): Threshold. Defaults to 0.5.

    Returns:
        float: Center of mass.
    """
    y /= np.max(y)
    y -= thres
    y[y < 0] = 0
    y /= np.sum(y)
    com = np.dot(x, y)
    return com


def pulse_freq(
    pulse: np.ndarray, fftlen: int = 1000, sampling_rate: int = 10000, mean_subtract: bool = True
) -> Tuple[float, np.ndarray, np.ndarray]:
    """Calculate pulse frequency as center of mass of the pulse's amplitude spectrum.

    Args:
        pulse (np.ndarray): Waveform (shape [T,]).
        fftlen (int, optional): Sets freq resolution of the spectrum. Defaults to 1_000.
        sampling_rate (float, optional): Sample rate of the pulse, in Hz. Defaults to 10_000.
        mean_subtract (bool, optional): If true, removes f0 component by mean subtraction. Defaults to True.
    Returns:
        Tuple[float, np.ndarray, np.ndarray]: Center frequency,
                                              frequency and
                                              amplitude values of the pulse spectrum (cut off at 1000 Hz).
    """
    if mean_subtract:
        pulse -= np.mean(pulse)
    F = np.fft.rfftfreq(fftlen, 1 / sampling_rate)
    A = np.abs(np.fft.rfft(pulse, fftlen))
    idx = int(np.argmax(F > 1_000))
    center_freq = center_of_mass(F[:idx], A[:idx])
    return center_freq, F[:idx], A[:idx]


def get_pulseshapes(pulsecenters: List[int], song: np.ndarray, win_hw: int) -> np.ndarray:
    """Extract waveforms around `pulsecenters` from `song`.

    In case of multi-channel recordings, will return the waveform on the channel
    with the maximum absolute value within +/-`win_hw` around the each pulsecenter.

    Args:
        pulsecenters (List[int]): Location of each pulse center in `song`, in samples
        song (np.ndarray): Audio data ([samples, channels]).
        win_hw (int): Half-width of the waveform cut out around each pulse center, in samples.

    Returns:
        np.ndarray: Extracted waveforms [2 * win_hw, nb_centers]
    """
    pulseshapes = np.zeros((2 * win_hw, len(pulsecenters)))
    nb_channels = song.shape[1]
    for cnt, p in enumerate(pulsecenters):
        t0 = int(p - win_hw)
        t1 = int(p + win_hw)
        if t0 > 0 and t1 < song.shape[0]:
            if nb_channels > 1:
                tmp = song[t0:t1, :]
                loudest_channel = np.argmax(np.max(np.abs(tmp), axis=0))
                pulseshapes[:, cnt] = tmp[:, loudest_channel].copy()
            else:
                pulseshapes[:, cnt] = song[t0:t1, 0].copy()
    return pulseshapes



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import numpy as np
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer


class AdditiveNoise(Layer):
    """Adds Gaussian noise to the spectrogram."""

    def __init__(self, power: float = 0.1, random_gain: True = False, noise_type: str = "white", **kwargs):
        """Init.

        Args:
            power (float, optional): Standard deviation of the noise. Defaults to 0.1.
            random_gain (True, optional): If `True`, gain is sampled from `uniform(low=0.0, high=power)` in every batch. Defaults to False.
            noise_type (str, optional): Only supports white. Defaults to 'white'.
        """
        assert noise_type in ["white"]
        self.supports_masking = True
        self.power = power
        self.random_gain = random_gain
        self.noise_type = noise_type
        self.uses_learning_phase = True
        super(AdditiveNoise, self).__init__(**kwargs)

    def call(self, x):
        if self.random_gain:
            power = np.random.uniform(0.0, self.power)
        else:
            power = self.power
        noise_x = x + K.random_normal(shape=K.shape(x), mean=0.0, stddev=power)
        return K.in_train_phase(noise_x, x)

    def get_config(self):
        config = {"power": self.power, "random_gain": self.random_gain, "noise_type": self.noise_type}
        base_config = super(AdditiveNoise, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import sklearn.metrics
import numpy as np
import pandas as pd
from typing import Optional, Dict, Callable, Any
import logging
from . import predict, data, utils, models, io
from .event_utils import evaluate_eventtimes


# to segment_utils
def evaluate_segments(
    labels_test, labels_pred, class_names, confmat_as_pandas: bool = False, report_as_dict: bool = False, labels=None
):
    """

    Args:
        labels_test (List): [nb_samples,]
        labels_pred (List): [nb_samples,]
        class_names ([type]): [description]
        confmat_as_pandas (bool, optional): [description]. Defaults to False.
        report_as_dict (bool, optional): [description]. Defaults to False.
        labels ([type], optional): [description]. Defaults to None.

    Returns:
        conf_mat
        report
    """

    # ensure labels_test and labels_pred have same length
    min_len = min(len(labels_test), len(labels_pred))
    labels_test = labels_test[:min_len]
    labels_pred = labels_pred[:min_len]

    conf_mat = sklearn.metrics.confusion_matrix(labels_test, labels_pred)
    if confmat_as_pandas:
        conf_mat = pd.DataFrame(
            data=conf_mat, columns=["true " + p for p in class_names], index=["pred " + p for p in class_names]
        )
    if labels is None:
        labels = np.arange(len(class_names))
    report = sklearn.metrics.classification_report(
        labels_test, labels_pred, labels=labels, target_names=class_names, output_dict=report_as_dict, digits=3
    )
    return conf_mat, report


def evaluate_segment_timing(segment_labels_true, segment_labels_pred, samplerate: float, event_tol: float):
    """[summary]

    Args:
        segment_labels_true ([type]): [description]
        segment_labels_pred ([type]): [description]
        samplerate ([type]): Hz [description]
        event_tol ([type]): seconds [description]

    Returns:
        [type]: [description]
    """
    segment_onsets_true, segment_offsets_true = segment_timing(segment_labels_true, samplerate)
    segment_onsets_pred, segment_offsets_pred = segment_timing(segment_labels_pred, samplerate)

    # ensure evaluate_eventtimes returns nearest_predicted_onsets (nearest true event for each predicted event),
    # if not, rename var
    segment_onsets_report, _, _, nearest_predicted_onsets = evaluate_eventtimes(
        segment_onsets_true, segment_onsets_pred, samplerate, event_tol
    )
    segment_offsets_report, _, _, nearest_predicted_offsets = evaluate_eventtimes(
        segment_offsets_true, segment_offsets_pred, samplerate, event_tol
    )
    return segment_onsets_report, segment_offsets_report, nearest_predicted_onsets, nearest_predicted_offsets


# TODO: move to das.segment_utils
def segment_timing(labels, samplerate: float):
    """Get onset and offset time (in seconds) for each segment."""
    segment_onset_times = np.where(np.diff(labels) == 1)[0].astype(np.float) / samplerate  # explicit cast required?
    segment_offset_times = np.where(np.diff(labels) == -1)[0].astype(np.float) / samplerate
    return segment_onset_times, segment_offset_times


def evaluate_probabilities(
    x,
    y,
    model: Optional[models.keras.models.Model] = None,
    params: Optional[Dict] = None,
    model_savename: Optional[str] = None,
    verbose: int = 1,
):
    """[summary]

    evaluate_probabilities(x, y, model=keras_model, params=params_dict)
    evaluate_probabilities(x, y, model_savename=save_string) -> will load model and params

    Args:
        x ([type]): [description]
        y ([type]): [description]
        model (Union[models.keras.models.Model], optional): [description]. Defaults to None.
        params (Union[Dict], optional): [description]. Defaults to None.
        model_savename (Union[str], optional): [description]. Defaults to None.
        verbose (int, optional): [description]. Defaults to 1.

    Returns:
        [type]: [description]
    """
    # TODO if called w/o x and y, load dataset from params
    if model is None or params is None:
        if model_savename is not None:
            model, params = utils.load_model_and_params(model_savename)
        else:
            raise ValueError(
                f"Required: Either a model and params OR a model_savename so we can load model and params. But model={model}, params={params}, model_savename={model_savename}."
            )

    # do not prepend padding since we create y from the generator
    y_pred = predict.predict_probabilities(x, model=model, params=params, verbose=verbose, prepend_data_padding=False)

    eval_gen = data.AudioSequence(x, y, shuffle=False, **params)
    x, y = data.get_data_from_gen(eval_gen)

    return x, y, y_pred


def evaluate(savename: str, custom_objects: Optional[Dict[str, Callable]] = None, full_output: bool = True, verbose: int = 1):
    logging.info("Loading last best model.")
    model, params = utils.load_model_and_params(savename, custom_objects=custom_objects)
    logging.info(model.summary())

    logging.info(f"Loading data from {params['data_dir']}.")
    d = io.load(params["data_dir"], x_suffix=params["x_suffix"], y_suffix=params["y_suffix"])

    output: Dict[str, Any] = dict()
    if len(d["test"]["x"]) < params["nb_hist"]:
        logging.info("No test data - skipping final evaluation step.")
        if full_output:
            return None, None, output
        else:
            return None, None
    else:
        logging.info("predicting")
        x_test, y_test, y_pred = evaluate_probabilities(
            x=d["test"]["x"], y=d["test"]["y"], model=model, params=params, verbose=verbose
        )

        labels_test = predict.labels_from_probabilities(y_test)
        labels_pred = predict.labels_from_probabilities(y_pred)

        logging.info("evaluating")
        conf_mat, report = evaluate_segments(labels_test, labels_pred, params["class_names"], report_as_dict=True)
        logging.info(conf_mat)
        logging.info(report)
        output["x_test"] = x_test
        output["y_test"] = y_test
        output["y_pred"] = y_pred
        output["labels_test"] = labels_test
        output["labels_pred"] = labels_pred
        if full_output:
            return conf_mat, report, output
        else:
            return conf_mat, report
